#!/usr/bin/env python3

# FIXME:  Port this pipeline to 2.3.


# Copyright (c) 2015-2018 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from toast.mpi import MPI, finalize

import os
import sys
import time

import re
import argparse
import traceback

import pickle

import numpy as np

import toast
import toast.tod as tt
import toast.map as tm
import toast.todmap as ttm

import toast.qarray as qa
import toast.timing as timing

from toast.vis import set_backend

if tt.tidas_available:
    from toast.tod import tidas as tds

if tt.spt3g_available:
    from toast.tod import spt3g as s3g


def elapsed(mcomm, start, msg):
    mcomm.barrier()
    stop = MPI.Wtime()
    dur = stop - start
    if mcomm.rank == 0:
        print("{}: {:.3f} s".format(msg, dur), flush=True)
    return stop


def main():

    if MPI.COMM_WORLD.rank == 0:
        print("Running with {} processes".format(MPI.COMM_WORLD.size), flush=True)

    global_start = MPI.Wtime()

    parser = argparse.ArgumentParser(
        description="Read existing data and make a simple map.",
        fromfile_prefix_chars="@",
    )

    parser.add_argument(
        "--groupsize",
        required=False,
        type=int,
        default=0,
        help="size of processor groups used to distribute " "observations",
    )

    parser.add_argument(
        "--hwprpm",
        required=False,
        type=float,
        default=0.0,
        help="The rate (in RPM) of the HWP rotation",
    )

    parser.add_argument(
        "--samplerate",
        required=False,
        default=100.0,
        type=np.float,
        help="Detector sample rate (Hz)",
    )

    parser.add_argument(
        "--outdir", required=False, default="out", help="Output directory"
    )

    parser.add_argument(
        "--nside", required=False, type=int, default=64, help="Healpix NSIDE"
    )

    parser.add_argument(
        "--subnside",
        required=False,
        type=int,
        default=8,
        help="Distributed pixel sub-map NSIDE",
    )

    parser.add_argument(
        "--coord", required=False, default="E", help="Sky coordinate system [C,E,G]"
    )

    parser.add_argument(
        "--baseline",
        required=False,
        type=float,
        default=60.0,
        help="Destriping baseline length (seconds)",
    )

    parser.add_argument(
        "--noisefilter",
        required=False,
        default=False,
        action="store_true",
        help="Destripe with the noise filter enabled",
    )

    parser.add_argument(
        "--madam",
        required=False,
        default=False,
        action="store_true",
        help="If specified, use libmadam for map-making",
    )

    parser.add_argument(
        "--madampar", required=False, default=None, help="Madam parameter file"
    )

    parser.add_argument(
        "--polyorder",
        required=False,
        type=int,
        help="Polynomial order for the polyfilter",
    )

    parser.add_argument(
        "--wbin_ground",
        required=False,
        type=float,
        help="Ground template bin width [degrees]",
    )

    parser.add_argument(
        "--flush",
        required=False,
        default=False,
        action="store_true",
        help="Flush every print statement.",
    )

    parser.add_argument(
        "--tidas", required=False, default=None, help="Input TIDAS volume"
    )

    parser.add_argument(
        "--tidas_detgroup", required=False, default=None, help="TIDAS detector group"
    )

    parser.add_argument(
        "--spt3g", required=False, default=None, help="Input SPT3G data directory"
    )

    parser.add_argument(
        "--spt3g_prefix",
        required=False,
        default=None,
        help="SPT3G data frame file prefix",
    )

    parser.add_argument(
        "--common_flag_mask",
        required=False,
        default=0,
        type=np.uint8,
        help="Common flag mask",
    )

    parser.add_argument(
        "--debug",
        required=False,
        default=False,
        action="store_true",
        help="Write data distribution info and focalplane plot",
    )

    args = timing.add_arguments_and_parse(parser, timing.FILE(noquotes=True))
    # args = parser.parse_args(sys.argv)

    autotimer = timing.auto_timer("@{}".format(timing.FILE()))

    if (args.tidas is not None) and (args.spt3g is not None):
        raise RuntimeError("Cannot read two datasets!")

    if (args.tidas is None) and (args.spt3g is None):
        raise RuntimeError("No dataset specified!")

    if args.tidas is not None:
        if not tt.tidas_available:
            raise RuntimeError("TIDAS not found- cannot load")

    if args.spt3g is not None:
        if not tt.spt3g_available:
            raise RuntimeError("SPT3G not found- cannot load")

    groupsize = args.groupsize
    if groupsize == 0:
        groupsize = MPI.COMM_WORLD.size

    # Pixelization

    nside = args.nside
    npix = 12 * args.nside * args.nside
    subnside = args.subnside
    if subnside > nside:
        subnside = nside
    subnpix = 12 * subnside * subnside

    # This is the 2-level toast communicator.

    if MPI.COMM_WORLD.size % groupsize != 0:
        if MPI.COMM_WORLD.rank == 0:
            print(
                "WARNING:  process groupsize does not evenly divide into "
                "total number of processes",
                flush=True,
            )
    comm = toast.Comm(world=MPI.COMM_WORLD, groupsize=groupsize)

    # Create output directory

    mtime = MPI.Wtime()

    if comm.comm_world.rank == 0:
        if not os.path.isdir(args.outdir):
            os.makedirs(args.outdir)

    mtime = elapsed(comm.comm_world, mtime, "Creating output directory")

    # The distributed timestream data

    data = None

    if args.tidas is not None:
        if args.tidas_detgroup is None:
            raise RuntimeError("you must specify the detector group")
        data = tds.load_tidas(
            comm,
            comm.group_size,
            args.tidas,
            "r",
            args.tidas_detgroup,
            tds.TODTidas,
            group_dets=args.tidas_detgroup,
            distintervals="chunks",
        )

    if args.spt3g is not None:
        if args.spt3g_prefix is None:
            raise RuntimeError("you must specify the frame file prefix")
        data = s3g.load_spt3g(
            comm,
            comm.group_size,
            args.spt3g,
            args.spt3g_prefix,
            s3g.obsweight_spt3g,
            s3g.TOD3G,
        )

    mtime = elapsed(comm.comm_world, mtime, "Distribute data")

    # In debug mode, print out data distribution information

    if args.debug:
        handle = None
        if comm.comm_world.rank == 0:
            handle = open("{}_distdata.txt".format(args.outdir), "w")
        data.info(handle)
        if comm.comm_world.rank == 0:
            handle.close()
        mtime = elapsed(comm.comm_world, mtime, "Dumping debug data distribution")
        if comm.comm_world.rank == 0:
            outfile = "{}_focalplane.png".format(args.outdir)
            set_backend()
            # Just plot the dets from the first TOD
            temptod = data.obs[0]["tod"]
            # FIXME: change this once we store det info in the metadata.
            dfwhm = {x: 10.0 for x in temptod.detectors}
            tt.plot_focalplane(temptod.detoffset(), 10.0, 10.0, outfile, fwhm=dfwhm)
        comm.comm_world.barrier()
        mtime = elapsed(comm.comm_world, mtime, "Plotting debug focalplane")

    # Compute pointing matrix

    pointing = tt.OpPointingHpix(
        nside=args.nside, nest=True, mode="IQU", hwprpm=args.hwprpm
    )
    pointing.exec(data)

    mtime = elapsed(comm.comm_world, mtime, "Expand pointing")

    # Mapmaking.

    # FIXME:  We potentially have a different noise model for every
    # observation.  We need to have both spt3g and tidas format Noise
    # classes which read the information from disk.  Then the mapmaking
    # operators need to get these noise weights from each observation.
    detweights = {d: 1.0 for d in data.obs[0]["tod"].detectors}

    if not args.madam:
        if comm.comm_world.rank == 0:
            print("Not using Madam, will only make a binned map!", flush=True)

        # Filter data if desired

        if args.polyorder:
            polyfilter = tt.OpPolyFilter(
                order=args.polyorder, common_flag_mask=args.common_flag_mask
            )
            polyfilter.exec(data)
            mtime = elapsed(comm.comm_world, mtime, "Polynomial filtering")

        if args.wbin_ground:
            groundfilter = tt.OpGroundFilter(
                wbin=args.wbin_ground, common_flag_mask=args.common_flag_mask
            )
            groundfilter.exec(data)
            mtime = elapsed(comm.comm_world, mtime, "Ground template filtering")

        # Compute pixel space distribution

        lc = tm.OpLocalPixels()
        localpix = lc.exec(data)
        if localpix is None:
            raise RuntimeError(
                "Process {} has no hit pixels. Perhaps there are fewer "
                "detectors than processes in the group?".format(comm.comm_world.rank)
            )
        localsm = np.unique(np.floor_divide(localpix, subnpix))
        mtime = elapsed(comm.comm_world, mtime, "Compute local submaps")

        # construct distributed maps to store the covariance,
        # noise weighted map, and hits

        mtime = MPI.Wtime()
        invnpp = tm.DistPixels(
            comm=comm.comm_world,
            size=npix,
            nnz=6,
            dtype=np.float64,
            submap=subnpix,
            local=localsm,
        )
        hits = tm.DistPixels(
            comm=comm.comm_world,
            size=npix,
            nnz=1,
            dtype=np.int64,
            submap=subnpix,
            local=localsm,
        )
        zmap = tm.DistPixels(
            comm=comm.comm_world,
            size=npix,
            nnz=3,
            dtype=np.float64,
            submap=subnpix,
            local=localsm,
        )

        # compute the hits and covariance.

        invnpp.data.fill(0.0)
        hits.data.fill(0)

        build_invnpp = tm.OpAccumDiag(
            detweights=detweights,
            invnpp=invnpp,
            hits=hits,
            common_flag_mask=args.common_flag_mask,
        )
        build_invnpp.exec(data)

        invnpp.allreduce()
        hits.allreduce()
        mtime = elapsed(comm.comm_world, mtime, "Building hits and N_pp^-1")

        hits.write_healpix_fits("{}_hits.fits".format(args.outdir))
        invnpp.write_healpix_fits("{}_invnpp.fits".format(args.outdir))
        mtime = elapsed(comm.comm_world, mtime, "Writing hits and N_pp^-1")

        # invert it
        tm.covariance_invert(invnpp, 1.0e-3)
        mtime = elapsed(comm.comm_world, mtime, "Inverting N_pp^-1")

        invnpp.write_healpix_fits("{}_npp.fits".format(args.outdir))
        mtime = elapsed(comm.comm_world, mtime, "Writing N_pp")

        zmap.data.fill(0.0)
        build_zmap = tm.OpAccumDiag(
            zmap=zmap, detweights=detweights, common_flag_mask=args.common_flag_mask
        )
        build_zmap.exec(data)
        zmap.allreduce()
        mtime = elapsed(comm.comm_world, mtime, "Building noise weighted map")

        tm.covariance_apply(invnpp, zmap)
        mtime = elapsed(comm.comm_world, mtime, "Computing binned map")

        zmap.write_healpix_fits(os.path.join(args.outdir, "binned.fits"))
        mtime = elapsed(comm.comm_world, mtime, "Writing binned map")

    else:
        # Set up MADAM map making.

        pars = {}
        pars["temperature_only"] = "F"
        pars["force_pol"] = "T"
        pars["kfirst"] = "T"
        pars["concatenate_messages"] = "T"
        pars["write_map"] = "T"
        pars["write_binmap"] = "T"
        pars["write_matrix"] = "T"
        pars["write_wcov"] = "T"
        pars["write_hits"] = "T"
        pars["nside_cross"] = nside // 2
        pars["nside_submap"] = subnside

        if args.madampar is not None:
            pat = re.compile(r"\s*(\S+)\s*=\s*(\S+(\s+\S+)*)\s*")
            comment = re.compile(r"^#.*")
            with open(args.madampar, "r") as f:
                for line in f:
                    if comment.match(line) is None:
                        result = pat.match(line)
                        if result is not None:
                            key, value = result.group(1), result.group(2)
                            pars[key] = value

        pars["base_first"] = args.baseline
        pars["nside_map"] = nside
        if args.noisefilter:
            pars["kfilter"] = "T"
        else:
            pars["kfilter"] = "F"
        pars["fsample"] = args.samplerate

        madam = tm.OpMadam(
            params=pars, detweights=detweights, common_flag_mask=args.common_flag_mask
        )
        madam.exec(data)
        mtime = elapsed(comm.comm_world, mtime, "Madam mapmaking")

    comm.comm_world.barrier()
    stop = MPI.Wtime()
    dur = stop - global_start
    if comm.comm_world.rank == 0:
        print("Total Time:  {:.2f} seconds".format(dur), flush=True)
    return


if __name__ == "__main__":
    try:
        main()
        tman = timing.timing_manager()
        tman.report()
    except:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        lines = ["Proc {}: {}".format(MPI.COMM_WORLD.rank, x) for x in lines]
        print("".join(lines), flush=True)
        toast.raise_error(6)  # typical error code for SIGABRT
        MPI.COMM_WORLD.Abort(6)
    finalize()



pysm_sky_cmb = PySMSky(comm=None,
       pixel_indices=None,
       nside=NSIDE,
       units="uK_CMB",
       pysm_sky_config=["c1"]
)

pysm_sky_cmb.exec(local_maps, out="cmb", bandpasses={"ch0":(100,1),"ch1":(200,1)})

hp.mollview(local_maps["cmb_ch0"][0]-local_maps["cmb_ch1"][0])



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .mpi import MPITestCase

import os

import numpy as np

from ..tod import OpMemoryCounter, AnalyticNoise, OpSimNoise
from ..todmap import TODHpixSpiral

from ._helpers import (
    create_outdir,
    create_distdata,
    boresight_focalplane,
    uniform_chunks,
)


class OpMemoryCounterTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        # One observation per group
        self.data = create_distdata(self.comm, obs_per_group=1)

        # Detector properties.  We place one detector per process at the
        # boresight with evenly spaced polarization orientations.

        self.ndet = 4
        self.NET = 5.0
        self.rate = 20.0

        # Create detectors with default properties.
        (
            dnames,
            dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(self.ndet, samplerate=self.rate, net=self.NET)

        # Total samples per observation
        self.totsamp = 100000

        # Chunks
        chunks = uniform_chunks(self.totsamp, nchunk=self.data.comm.group_size)

        # Populate the observations

        tod = TODHpixSpiral(
            self.data.comm.comm_group,
            dquat,
            self.totsamp,
            detranks=1,
            firsttime=0.0,
            rate=self.rate,
            nside=512,
            sampsizes=chunks,
        )

        # construct an analytic noise model

        nse = AnalyticNoise(
            rate=drate,
            fmin=dfmin,
            detectors=dnames,
            fknee=dfknee,
            alpha=dalpha,
            NET=dnet,
        )

        self.data.obs[0]["tod"] = tod
        self.data.obs[0]["noise"] = nse

    def test_counter(self):
        ob = self.data.obs[0]
        tod = ob["tod"]
        # Ensure timestamps are cached before simulating noise
        blah = tod.local_times()
        del blah

        counter = OpMemoryCounter(silent=True)

        tot_old = counter.exec(self.data)

        # generate timestreams
        op = OpSimNoise()
        op.exec(self.data)

        tot_new = counter.exec(self.data)

        expected = self.data.comm.ngroups * self.totsamp * self.ndet * 8

        np.testing.assert_equal(tot_new - tot_old, expected)

        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .mpi import MPITestCase

import os

import numpy as np
import numpy.testing as nt

from ..dist import distribute_uniform, distribute_discrete, Data
from ..mpi import Comm, MPI

from ._helpers import create_outdir, create_distdata


class DataTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        # Create one observation per group.

        self.data = create_distdata(self.comm, obs_per_group=1)

        self.ntask = 24
        self.sizes1 = [
            29218,
            430879,
            43684,
            430338,
            36289,
            437553,
            37461,
            436200,
            41249,
            432593,
            42467,
            431195,
            35387,
            438274,
            36740,
            436741,
            40663,
            432999,
            42015,
            431285,
            35297,
            438004,
            37010,
            436291,
            41114,
            432186,
            42828,
            430293,
            36243,
            436697,
            38318,
            434802,
            42602,
            430338,
            44676,
            428264,
            38273,
            434306,
            40708,
            432051,
            45308,
            427452,
            36695,
            435884,
            41520,
            430879,
            44090,
            428309,
            38273,
            434126,
            40843,
            431375,
        ]
        self.totsamp1 = np.sum(self.sizes1)

        self.sizes2 = [(int(3600 * 169.7)) for i in range(8640)]
        self.totsamp2 = np.sum(self.sizes2)

    def test_construction(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        dist_uni1 = distribute_uniform(self.totsamp1, self.ntask)
        # with open("test_uni_{}".format(self.comm.rank), "w") as f:
        #     for d in dist_uni:
        #         f.write("uniform:  {} {}\n".format(d[0], d[1]))
        n1 = np.sum(np.array(dist_uni1)[:, 1])
        assert n1 == self.totsamp1

        n = self.totsamp1
        breaks = [n // 2 + 1000, n // 4 - 1000000, n // 2 + 1000, (3 * n) // 4]
        dist_uni2 = distribute_uniform(self.totsamp1, self.ntask, breaks=breaks)

        n2 = np.sum(np.array(dist_uni2)[:, 1])
        assert n2 == self.totsamp1

        for offset, nsamp in dist_uni2:
            for brk in breaks:
                if brk > offset and brk < offset + nsamp:
                    raise Exception(
                        "Uniform data distribution did not honor the breaks"
                    )

        dist_disc1 = distribute_discrete(self.sizes1, self.ntask)
        # with open("test_disc_{}".format(self.comm.rank), "w") as f:
        #     for d in dist_disc:
        #         f.write("discrete:  {} {}\n".format(d[0], d[1]))

        n = np.sum(np.array(dist_disc1)[:, 1])
        assert n == len(self.sizes1)

        n = len(self.sizes1)
        breaks = [n // 2, n // 4, n // 2, (3 * n) // 4]
        dist_disc2 = distribute_discrete(self.sizes1, self.ntask, breaks=breaks)

        n = np.sum(np.array(dist_disc2)[:, 1])
        assert n == len(self.sizes1)

        for offset, nchunk in dist_disc2:
            for brk in breaks:
                if brk > offset and brk < offset + nchunk:
                    raise Exception(
                        "Discrete data distribution did not honor the breaks"
                    )

        handle = None
        if rank == 0:
            handle = open(os.path.join(self.outdir, "out_test_construct_info"), "w")
        self.data.info(handle=handle)
        if rank == 0:
            handle.close()

        dist_disc3 = distribute_discrete(self.sizes2, 384)

        if rank == 0:
            with open(
                os.path.join(self.outdir, "dist_discrete_8640x384.txt"), "w"
            ) as f:
                indx = 0
                for d in dist_disc3:
                    f.write("{:04d} = ({}, {})\n".format(indx, d[0], d[1]))
                    indx += 1
        return

    def test_split(self):
        data = Data(self.data.comm)
        data.obs.append({"site": "Atacama", "season": 1})
        data.obs.append({"site": "Atacama", "season": 2})
        data.obs.append({"site": "Atacama", "season": 3})
        data.obs.append({"site": "Pole", "season": 1})
        data.obs.append({"site": "Pole", "season": 2})
        data.obs.append({"site": "Pole", "season": 3})

        datasplit_site = data.split("site")
        datasplit_season = data.split("season")

        nt.assert_equal(len(datasplit_site), 2)
        nt.assert_equal(len(datasplit_season), 3)

        # Verify that the observations are shared

        sum1 = 0
        for value, site_data in datasplit_site:
            for obs in site_data.obs:
                assert "var1" not in obs
                obs["var1"] = 1
                sum1 += 1

        sum2 = 0
        for value, season_data in datasplit_season:
            for obs in season_data.obs:
                sum2 += obs["var1"]

        nt.assert_equal(sum1, sum2)
        return

    def test_none(self):
        # test that Comm with None argument returns a None communicator
        if MPI is None:
            comm = Comm(None)
        else:
            comm = Comm(MPI.COMM_SELF)
        assert comm.comm_world is None



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import argparse

import numpy as np

from toast.timing import function_timer, Timer
from toast.utils import Logger, Environment

from ..tod import OpPolyFilter, OpPolyFilter2D
from ..todmap import OpGroundFilter

#
# Polynomial filter
#


def add_polyfilter2D_args(parser):
    """Add the polynomial filter arguments to argparser"""
    parser.add_argument(
        "--polyfilter2D",
        required=False,
        default=False,
        action="store_true",
        help="Apply 2D polynomial filter",
        dest="apply_polyfilter2D",
    )
    parser.add_argument(
        "--no-polyfilter2D",
        required=False,
        action="store_false",
        help="Do not apply 2D polynomial filter",
        dest="apply_polyfilter2D",
    )
    parser.set_defaults(apply_polyfilter2D=False)

    parser.add_argument(
        "--poly-order2D",
        required=False,
        default=0,
        type=np.int,
        help="Polynomial order for the 2D polyfilter",
    )
    # Common flag mask may already be added
    try:
        parser.add_argument(
            "--common-flag-mask",
            required=False,
            default=1,
            type=np.uint8,
            help="Common flag mask",
        )
    except argparse.ArgumentError:
        pass
    return


def add_polyfilter_args(parser):
    """Add the polynomial filter arguments to argparser"""
    parser.add_argument(
        "--polyfilter",
        required=False,
        default=False,
        action="store_true",
        help="Apply polynomial filter",
        dest="apply_polyfilter",
    )
    parser.add_argument(
        "--no-polyfilter",
        required=False,
        action="store_false",
        help="Do not apply polynomial filter",
        dest="apply_polyfilter",
    )
    parser.set_defaults(apply_polyfilter=False)

    parser.add_argument(
        "--poly-order",
        required=False,
        default=0,
        type=np.int,
        help="Polynomial order for the polyfilter",
    )
    # Common flag mask may already be added
    try:
        parser.add_argument(
            "--common-flag-mask",
            required=False,
            default=1,
            type=np.uint8,
            help="Common flag mask",
        )
    except argparse.ArgumentError:
        pass
    return


@function_timer
def apply_polyfilter2D(args, comm, data, cache_name=None, verbose=True):
    """Apply the 2D polynomial filter to data under `cache_name`."""
    if not args.apply_polyfilter2D:
        return
    log = Logger.get()
    timer = Timer()
    timer.start()
    if comm.world_rank == 0 and verbose:
        log.info("2D Polyfiltering signal")
    polyfilter = OpPolyFilter2D(
        order=args.poly_order2D, name=cache_name, common_flag_mask=args.common_flag_mask
    )
    polyfilter.exec(data)
    if comm.comm_world is not None:
        comm.comm_world.barrier()
    if comm.world_rank == 0 and verbose:
        timer.report_clear("2D Polynomial filtering")
    return


#
# Ground filter
#


@function_timer
def apply_polyfilter(args, comm, data, cache_name=None, verbose=True):
    """Apply the polynomial filter to data under `cache_name`."""
    if not args.apply_polyfilter:
        return
    log = Logger.get()
    timer = Timer()
    timer.start()
    if comm.world_rank == 0 and verbose:
        log.info("Polyfiltering signal")
    polyfilter = OpPolyFilter(
        order=args.poly_order, name=cache_name, common_flag_mask=args.common_flag_mask
    )
    polyfilter.exec(data)
    if comm.comm_world is not None:
        comm.comm_world.barrier()
    if comm.world_rank == 0 and verbose:
        timer.report_clear("Polynomial filtering")
    return


#
# Ground filter
#


def add_groundfilter_args(parser):
    """Add the ground filter arguments to argparser"""
    parser.add_argument(
        "--groundfilter",
        required=False,
        default=False,
        action="store_true",
        help="Apply ground filter",
        dest="apply_groundfilter",
    )
    parser.add_argument(
        "--no-groundfilter",
        required=False,
        action="store_false",
        help="Do not apply ground filter",
        dest="apply_groundfilter",
    )
    parser.set_defaults(apply_groundfilter=False)

    parser.add_argument(
        "--ground-order",
        required=False,
        default=0,
        type=np.int,
        help="Ground template order",
    )
    # Common flag mask may already be added
    try:
        parser.add_argument(
            "--common-flag-mask",
            required=False,
            default=1,
            type=np.uint8,
            help="Common flag mask",
        )
    except argparse.ArgumentError:
        pass
    return


@function_timer
def apply_groundfilter(args, comm, data, cache_name=None, verbose=True):
    if not args.apply_groundfilter:
        return
    log = Logger.get()
    timer = Timer()
    timer.start()
    if comm.world_rank == 0 and verbose:
        log.info("Ground-filtering signal")
    groundfilter = OpGroundFilter(
        filter_order=args.ground_order,
        name=cache_name,
        common_flag_mask=args.common_flag_mask,
    )
    groundfilter.exec(data)
    if comm.comm_world is not None:
        comm.comm_world.barrier()
    if comm.world_rank == 0 and verbose:
        timer.report_clear("Ground filtering")
    return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

from .. import qarray as qa


def cartesian_to_quat(offsets):
    """Convert cartesian angle offsets and rotation into quaternions.

    Focalplane geometries are often described in terms of wafer locations or
    separations given in simple X/Y angle offsets from a center point.
    this helper function converts such parameters into a quaternion describing
    the rotation.

    Args:
        offsets (list of arrays):  each item of the list has 3 elements for
            the X / Y angle offsets in degrees and the rotation in degrees
            about the Z axis.

    Returns:
        (list): list of quaternions for each item in the input list.

    """
    centers = list()
    zaxis = np.array([0, 0, 1], dtype=np.float64)
    for off in offsets:
        angrot = qa.rotation(zaxis, off[2] * np.pi / 180.0)
        wx = off[0] * np.pi / 180.0
        wy = off[1] * np.pi / 180.0
        wz = np.sqrt(1.0 - (wx * wx + wy * wy))
        wdir = np.array([wx, wy, wz])
        posrot = qa.from_vectors(zaxis, wdir)
        centers.append(qa.mult(posrot, angrot))
    return centers


def hex_nring(npix):
    """
    For a hexagonal layout with a given number of pixels, return the
    number of rings.
    """
    test = npix - 1
    nrings = 1
    while (test - 6 * nrings) >= 0:
        test -= 6 * nrings
        nrings += 1
    if test != 0:
        raise RuntimeError(
            "{} is not a valid number of pixels for a hexagonal layout".format(npix)
        )
    return nrings


def hex_row_col(npix, pix):
    """
    For a hexagonal layout, indexed in a "spiral" scheme (see hex_layout),
    this function returnes the "row" and "column" of a pixel.
    The row is zero along the main vertex-vertex axis, and is positive
    or negative above / below this line of pixels.
    """
    if pix >= npix:
        raise ValueError("pixel value out of range")
    test = npix - 1
    nrings = 1
    while (test - 6 * nrings) >= 0:
        test -= 6 * nrings
        nrings += 1
    if pix == 0:
        row = 0
        col = nrings - 1
    else:
        test = pix - 1
        ring = 1
        while (test - 6 * ring) >= 0:
            test -= 6 * ring
            ring += 1
        sector = int(test / ring)
        steps = np.mod(test, ring)
        coloff = nrings - ring - 1
        if sector == 0:
            row = steps
            col = coloff + 2 * ring - steps
        elif sector == 1:
            row = ring
            col = coloff + ring - steps
        elif sector == 2:
            row = ring - steps
            col = coloff
        elif sector == 3:
            row = -steps
            col = coloff
        elif sector == 4:
            row = -ring
            col = coloff + steps
        elif sector == 5:
            row = -ring + steps
            col = coloff + ring + steps
    return (row, col)


def hex_pol_angles_qu(npix, offset=0.0):
    """Generates a vector of detector polarization angles.

    The returned angles can be used to construct a hexagonal detector layout.
    This scheme alternates pixels between 0/90 and +/- 45 degrees.

    Args:
        npix (int): the number of pixels locations in the hexagon.
        offset (float): the constant angle offset in degrees to apply.

    Returns:
        (array):  The detector polarization angles.

    """
    pol = np.zeros(npix, dtype=np.float64)
    for pix in range(npix):
        # get the row / col of the pixel
        row, col = hex_row_col(npix, pix)
        if np.mod(col, 2) == 0:
            pol[pix] = 0.0 + offset
        else:
            pol[pix] = 45.0 + offset
    return pol


def hex_pol_angles_radial(npix, offset=0.0):
    """Generates a vector of detector polarization angles.

    The returned angles can be used to construct a hexagonal detector layout.
    This scheme orients the bolometer along the radial direction of the
    hexagon.

    Args:
        npix (int): the number of pixels locations in the hexagon.
        offset (float): the constant angle offset in degrees to apply.

    Returns:
        (array):  The detector polarization angles.

    """
    sixty = np.pi / 3.0
    thirty = np.pi / 6.0
    pol = np.zeros(npix, dtype=np.float64)
    pol[0] = 0.0
    for pix in range(1, npix):
        # find ring for this pix
        test = pix - 1
        ring = 1
        while (test - 6 * ring) >= 0:
            test -= 6 * ring
            ring += 1
        sectors = int(test / ring)
        sectorsteps = np.mod(test, ring)
        midline = 0.5 * np.sqrt(3) * float(ring)
        edgedist = float(sectorsteps) - 0.5 * float(ring)
        relang = np.arctan2(edgedist, midline)
        pol[pix] = (sectors * sixty + thirty + relang) * 180.0 / np.pi + offset
    return pol


def hex_layout(
    npix, angwidth, prefix, suffix, pol, center=np.array([0, 0, 0, 1], dtype=np.float64)
):
    """Return detectors in a hexagon layout.

    This maps the physical positions of pixels into angular positions
    from the hexagon center.  The X axis in the hexagon frame is along
    the vertex-to-opposite-vertex direction.  The Y axis is along
    flat-to-opposite-flat direction.  The origin is at the center of
    the wafer.  For example::

        Y ^             O O O
        |              O O O O
        |             O O + O O
        +--> X         O O O O
                        O O O

    Each pixel is numbered 1..npix and each detector is named by the
    prefix, the pixel number, and the suffix.  The first pixel is at
    the center, and then the pixels are numbered moving outward in rings.

    The extent of the hexagon is directly specified by the angwidth
    parameter.  These, along with the npix parameter, constrain the packing
    locations of the pixel centers.

    Args:
        npix (int): number of pixels packed onto wafer.
        angwidth (float): the angle (in degrees) subtended by the width.
        prefix (str): the detector name prefix.
        suffix (str): the detector name suffix.
        pol (ndarray): 1D array of detector polarization angles.  The
            rotation is applied to the hexagon center prior to rotation
            to the pixel location.
        center (ndarray): quaternion offset of the center of the layout.

    Returns:
        (dict) A dictionary keyed on detector name, with each value itself a
            dictionary of detector properties.

    """
    zaxis = np.array([0, 0, 1], dtype=np.float64)
    nullquat = np.array([0, 0, 0, 1], dtype=np.float64)
    sixty = np.pi / 3.0
    thirty = np.pi / 6.0
    rtthree = np.sqrt(3.0)
    rtthreebytwo = 0.5 * rtthree

    angwidth = angwidth * np.pi / 180.0

    # compute the diameter (vertex to vertex width)
    angdiameter = angwidth / np.cos(thirty)

    # find the angular packing size of one detector
    test = npix - 1
    nrings = 1
    while (test - 6 * nrings) >= 0:
        test -= 6 * nrings
        nrings += 1
    pixdiam = angdiameter / (2 * nrings - 1)

    # convert pol vector to radians
    pol *= np.pi / 180.0

    # number of digits for pixel indexing

    ndigit = 0
    test = npix
    while test > 0:
        test = test // 10
        ndigit += 1

    nameformat = "{{}}{{:0{}d}}{{}}".format(ndigit)

    # compute positions of all detectors

    dets = {}

    for pix in range(npix):
        dname = nameformat.format(prefix, pix, suffix)

        polrot = qa.rotation(zaxis, pol[pix])

        # center pixel has no offset
        pixrot = nullquat

        if pix != 0:
            # Not at the center, find ring for this pix
            test = pix - 1
            ring = 1
            while (test - 6 * ring) >= 0:
                test -= 6 * ring
                ring += 1
            sectors = int(test / ring)
            sectorsteps = np.mod(test, ring)

            # Convert angular steps around the ring into the angle and distance
            # in polar coordinates.  Each "sector" of 60 degrees is essentially
            # an equilateral triangle, and each step is equally spaced along the
            # the edge opposite the vertex:
            #
            #          O
            #         O O (step 2)
            #        O   O (step 1)
            #       X O O O (step 0)
            #
            # For a given ring, "R" (center is R=0), there are R steps along
            # the sector edge.  The line from the origin to the opposite edge
            # that bisects this triangle has length R*sqrt(3)/2.  For each
            # equally-spaced step, we use the right triangle formed with this
            # bisection line to compute the angle and radius within this sector.

            # the distance from the origin to the midpoint of the opposite side.
            midline = rtthreebytwo * float(ring)

            # the distance along the opposite edge from the midpoint (positive
            # or negative)
            edgedist = float(sectorsteps) - 0.5 * float(ring)

            # the angle relative to the midpoint line (positive or negative)
            relang = np.arctan2(edgedist, midline)

            # total angle is based on number of sectors we have and the angle
            # within the final sector.
            pixang = sectors * sixty + thirty + relang

            pixdist = rtthreebytwo * pixdiam * float(ring) / np.cos(relang)

            pixx = np.sin(pixdist) * np.cos(pixang)
            pixy = np.sin(pixdist) * np.sin(pixang)
            pixz = np.cos(pixdist)
            pixdir = np.array([pixx, pixy, pixz], dtype=np.float64)
            norm = np.sqrt(np.dot(pixdir, pixdir))
            pixdir /= norm

            pixrot = qa.from_vectors(zaxis, pixdir)

        dprops = {}
        dprops["quat"] = qa.mult(center, qa.mult(pixrot, polrot))
        dprops["polangle_deg"] = pol[pix]

        dets[dname] = dprops

    return dets


def rhomb_dim(npix):
    """
    For a rhombus layout, return the dimension of one side.
    """
    dim = int(np.sqrt(float(npix)))
    if dim ** 2 != npix:
        raise ValueError("number of pixels for a rhombus wafer must be square")
    return dim


def rhomb_row_col(npix, pix):
    """
    For a rhombus layout, indexed from top to bottom (see rhombus_layout),
    this function returnes the "row" and "column" of a pixel.  The column
    starts at zero on the left hand side of a row.
    """
    if pix >= npix:
        raise ValueError("pixel value out of range")
    dim = rhomb_dim(npix)
    col = pix
    rowcnt = 1
    row = 0
    while (col - rowcnt) >= 0:
        col -= rowcnt
        row += 1
        if row >= dim:
            rowcnt -= 1
        else:
            rowcnt += 1
    return (row, col)


def rhomb_pol_angles_qu(npix, offset=0.0):
    """Generates a vector of detector polarization angles.

    The returned angles can be used to construct a rhombus detector layout.
    This scheme alternates pixels between 0/90 and +/- 45 degrees.

    Args:
        npix (int): the number of pixels locations in the rhombus.
        offset (float): the constant angle offset in degrees to apply.

    Returns:
        (array): The detector polarization angles.

    """
    pol = np.zeros(npix, dtype=np.float64)
    for pix in range(npix):
        # get the row / col of the pixel
        row, col = rhomb_row_col(npix, pix)
        if np.mod(col, 2) == 0:
            pol[pix] = 45.0 + offset
        else:
            pol[pix] = 0.0 + offset
    return pol


def rhombus_layout(
    npix,
    angwidth,
    prefix,
    suffix,
    polang,
    center=np.array([0, 0, 0, 1], dtype=np.float64),
):
    """Return detectors in a rhombus layout.

    This particular rhombus geometry is essentially a third of a
    hexagon.  In other words the aspect ratio of the rhombus is
    constrained to have the long dimension be sqrt(3) times the short
    dimension.

    This function maps the physical positions of pixels into angular
    positions from the rhombus center.  The X axis is along the short
    direction.  The Y axis is along longer direction.  The origin is
    at the center of the rhombus.  For example::

                          O
        Y ^              O O
        |               O O O
        |              O O O O
        +--> X          O O O
                         O O
                          O

    Each pixel is numbered 1..npix and each detector is named by the
    prefix, the pixel number, and the suffix.  The first pixel is at the
    "top", and then the pixels are numbered moving downward and left to
    right.

    The extent of the rhombus is directly specified by the angwidth parameter.
    This, along with the npix parameter, constrain the packing locations of
    the pixel centers.

    Args:
        npix (int): number of pixels packed onto wafer.
        angwidth (float): the angle (in degrees) subtended by the short
            dimension.
        prefix (str): the detector name prefix.
        suffix (str): the detector name suffix.
        polang (ndarray): 1D array of detector polarization angles.  The
            rotation is applied to the hexagon center prior to rotation
            to the pixel location.
        center (ndarray): quaternion offset of the center of the layout.

    Returns:
        (dict):  A dictionary keyed on detector name, with each value itself a
            dictionary of detector properties.

    """
    zaxis = np.array([0, 0, 1], dtype=np.float64)
    nullquat = np.array([0, 0, 0, 1], dtype=np.float64)
    rtthree = np.sqrt(3.0)

    angwidth = angwidth * np.pi / 180.0
    dim = rhomb_dim(npix)

    # compute the height
    angheight = rtthree * angwidth

    # find the angular packing size of one detector
    pixdiam = angwidth / dim

    # convert pol vector to radians
    pol = polang * np.pi / 180.0

    # number of digits for pixel indexing

    ndigit = 0
    test = npix
    while test > 0:
        test = test // 10
        ndigit += 1

    nameformat = "{{}}{{:0{}d}}{{}}".format(ndigit)

    # compute positions of all detectors

    dets = {}

    for pix in range(npix):
        dname = nameformat.format(prefix, pix, suffix)

        polrot = qa.rotation(zaxis, pol[pix])

        pixrow, pixcol = rhomb_row_col(npix, pix)

        rowang = 0.5 * rtthree * ((dim - 1) - pixrow) * pixdiam
        relrow = pixrow
        if pixrow >= dim:
            relrow = (2 * dim - 2) - pixrow
        colang = (float(pixcol) - float(relrow) / 2.0) * pixdiam
        distang = np.sqrt(rowang ** 2 + colang ** 2)
        zang = np.cos(distang)
        pixdir = np.array([colang, rowang, zang], dtype=np.float64)
        norm = np.sqrt(np.dot(pixdir, pixdir))
        pixdir /= norm

        pixrot = qa.from_vectors(zaxis, pixdir)

        dprops = {}
        dprops["quat"] = qa.mult(center, qa.mult(pixrot, polrot))

        dets[dname] = dprops

    return dets


def plot_focalplane(
    dets, width, height, outfile, fwhm=None, facecolor=None, polcolor=None, labels=None
):
    """Visualize a dictionary of detectors.

    This makes a simple plot of the detector positions on the projected
    focalplane.

    To avoid python overhead in large MPI jobs, we place the matplotlib
    import inside this function, so that it is only imported when the
    function is actually called.

    If the detector dictionary contains a key "fwhm", that will be assumed
    to be in arcminutes.  Otherwise a nominal value is used.

    If the detector dictionary contains a key "viscolor", then that color
    will be used.

    Args:
        dets (dict): dictionary of detector quaternions.
        width (float): width of plot in degrees.
        height (float): height of plot in degrees.
        outfile (str): output PNG path.  If None, then matplotlib will be
            used for inline plotting.
        fwhm (dict): dictionary of detector beam FWHM in arcminutes, used
            to draw the circles to scale.
        facecolor (dict): dictionary of color values for the face of each
            detector circle.
        polcolor (dict): dictionary of color values for the polarization
            arrows.
        labels (dict): plot this text in the center of each pixel.

    Returns:
        None

    """
    if outfile is not None:
        import matplotlib
        import warnings

        # Try to force matplotlib to not use any Xwindows backend.
        warnings.filterwarnings("ignore")
        matplotlib.use("Agg")
    import matplotlib.pyplot as plt

    xfigsize = int(5 * width)
    yfigsize = int(5 * height)
    figdpi = 75

    # Compute the font size to use for detector labels
    fontpix = 0.2 * figdpi
    fontpt = int(0.75 * fontpix)

    fig = plt.figure(figsize=(xfigsize, yfigsize), dpi=figdpi)
    ax = fig.add_subplot(1, 1, 1)

    half_width = 0.5 * width
    half_height = 0.5 * height
    ax.set_xlabel("Degrees", fontsize="large")
    ax.set_ylabel("Degrees", fontsize="large")
    ax.set_xlim([-half_width, half_width])
    ax.set_ylim([-half_height, half_height])

    xaxis = np.array([1.0, 0.0, 0.0], dtype=np.float64)
    yaxis = np.array([0.0, 1.0, 0.0], dtype=np.float64)
    zaxis = np.array([0.0, 0.0, 1.0], dtype=np.float64)

    for d, quat in dets.items():

        # radius in degrees
        detradius = 0.5 * 5.0 / 60.0
        if fwhm is not None:
            detradius = 0.5 * fwhm[d] / 60.0

        # rotation from boresight
        rdir = qa.rotate(quat, zaxis).flatten()
        ang = np.arctan2(rdir[1], rdir[0])

        orient = qa.rotate(quat, xaxis).flatten()
        polang = np.arctan2(orient[1], orient[0])

        mag = np.arccos(rdir[2]) * 180.0 / np.pi
        xpos = mag * np.cos(ang)
        ypos = mag * np.sin(ang)

        detface = "none"
        if facecolor is not None:
            detface = facecolor[d]

        circ = plt.Circle((xpos, ypos), radius=detradius, fc=detface, ec="k")
        ax.add_artist(circ)

        ascale = 2.0

        xtail = xpos - ascale * detradius * np.cos(polang)
        ytail = ypos - ascale * detradius * np.sin(polang)
        dx = ascale * 2.0 * detradius * np.cos(polang)
        dy = ascale * 2.0 * detradius * np.sin(polang)

        detcolor = "black"
        if polcolor is not None:
            detcolor = polcolor[d]

        ax.arrow(
            xtail,
            ytail,
            dx,
            dy,
            width=0.1 * detradius,
            head_width=0.3 * detradius,
            head_length=0.3 * detradius,
            fc=detcolor,
            ec=detcolor,
            length_includes_head=True,
        )

        if labels is not None:
            xsgn = 1.0
            if dx < 0.0:
                xsgn = -1.0
            labeloff = 0.05 * xsgn * fontpix * len(labels[d]) / figdpi
            ax.text(
                (xtail + 1.1 * dx + labeloff),
                (ytail + 1.1 * dy),
                labels[d],
                color="k",
                fontsize=fontpt,
                horizontalalignment="center",
                verticalalignment="center",
                bbox=dict(fc="w", ec="none", pad=1, alpha=1.0),
            )

    if outfile is None:
        plt.show()
    else:
        plt.savefig(outfile)
        plt.close()
    return fig



##
# Copyright (c) 2017-2020, all rights reserved.  Use of this source code
# is governed by a BSD license that can be found in the top-level
# LICENSE file.
##
"""Parallel shared memory tools.

This package contains tools for using synchronized shared memory across nodes
and implementing communicator-wide MUTEX locks.

"""

__version__ = "0.1.0"

# Namespace imports

from .shmem import MPIShared
from .locking import MPILock

# from .test import run as test



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

from astropy.io import fits

from ..op import Operator

from ..timing import function_timer

from .tod_math import calibrate

from ..utils import Logger


def write_calibration_file(filename, gain):
    """Write gains to a FITS file in the standard TOAST format

    Args:
        filename (string): output filename, overwritten by default
        gain (dict):  Dictionary, key "TIME" has the common timestamps,
            other keys are channel names their values are the gains

    """
    log = Logger.get()
    detectors = list(gain.keys())
    detectors.remove("TIME")

    hdus = [
        fits.PrimaryHDU(),
        fits.BinTableHDU.from_columns(
            [
                fits.Column(
                    name="DETECTORS",
                    array=detectors,
                    unit="",
                    format="{0}A".format(max([len(x) for x in detectors])),
                )
            ]
        ),
    ]
    hdus[1].header["EXTNAME"] = "DETECTORS"

    cur_hdu = fits.BinTableHDU.from_columns(
        [fits.Column(name="TIME", array=gain["TIME"], unit="s", format="1D")]
    )
    cur_hdu.header["EXTNAME"] = "TIME"
    hdus.append(cur_hdu)

    gain_table = np.zeros(
        (len(detectors), len(gain["TIME"])), dtype=gain[detectors[0]].dtype
    )
    for i_det, det in enumerate(detectors):
        gain_table[i_det, :] = gain[det]

    gainhdu = fits.ImageHDU(gain_table)
    gainhdu.header["EXTNAME"] = "GAINS"
    hdus.append(gainhdu)

    fits.HDUList(hdus).writeto(filename, overwrite=True)

    log.info("Gains written to file {}".format(filename))
    return


class OpApplyGain(Operator):
    """Operator which applies gains to timelines.

    Args:
        gain (dict):  Dictionary, key "TIME" has the common timestamps,
            other keys are channel names their values are the gains
        name (str):  Name of the output signal cache object will be
            <name_in>_<detector>.  If the object exists, it is used as
            input.  Otherwise signal is read using the tod read method.

    """

    def __init__(self, gain, name=None):
        self._gain = gain
        self._name = name
        # Call the parent class constructor
        super().__init__()

    @function_timer
    def exec(self, data):
        """Apply the gains.

        Args:
            data (toast.Data): The distributed data.

        """
        for obs in data.obs:

            tod = obs["tod"]

            for det in tod.local_dets:

                # Cache the output signal
                ref = tod.local_signal(det, self._name)
                obs_times = tod.read_times()

                calibrate(
                    obs_times,
                    ref,
                    self._gain["TIME"],
                    self._gain[det],
                    order=0,
                    inplace=True,
                )

                assert np.isnan(ref).sum() == 0, "The signal timestream includes NaN"

                del ref

        return



"""Helper tools common to all notebooks.
"""

import os
import numpy as np
import re

from toast.tod import hex_pol_angles_radial, hex_pol_angles_qu, hex_layout


def check_nersc(reservation=None, repo=None):
    """Check if we are running at NERSC.
    
    If we are at NERSC, select the account to use for batch jobs.
    
    Args:
        reservation (str):  Attempt to use this reservation for slurm jobs.
        repo (str):  In the case of multiple repos, use this one.
        
    Returns:
        (tuple):  The (host, repo, reservation) configured.

    """
    nersc_host = None
    nersc_repo = None
    nersc_resv = None
    if "NERSC_HOST" in os.environ:
        # We are at NERSC
        nersc_host = os.environ["NERSC_HOST"]
        # Be kind to the login node
        os.environ["OMP_NUM_THREADS"] = "4"
        import subprocess as sp
        repos = sp.check_output(
            "getnim -U $(whoami) | awk '{print $1}'", 
            shell=True,
            universal_newlines=True
        ).split()
        print(
            "Running on NERSC machine '{}'\n  with access to repos: {}".format(
                nersc_host, ", ".join(repos)
            )
        )
        if repo is not None:
            if repo in repos:
                nersc_repo = repo
                print("Using requested repo {}".format(repo))
            else:
                print("Requested repo {} not in list of enabled repos".format(repo))
        if nersc_repo is None:
            nersc_repo = repos[0]
            print("Using default repo {}".format(nersc_repo))
        if reservation is not None:
            # We would like to use a reservation
            checkres = sp.check_output(
                "scontrol show reservation {}".format(reservation), 
                shell=True,
                universal_newlines=True
            ).split()
            # Does this reservation even exist?
            if re.match(r".*not found.*", checkres[0]) is not None:
                print(
                    "Reservation '{}' does not exist or is expired".format(reservation)
                )
            else:
                startiso = None
                stopiso = None
                fullres = " ".join(checkres)
                startmat = re.match(r".*StartTime=(\S*).*", fullres)
                stopmat = re.match(r".*EndTime=(\S*).*", fullres)
                if startmat is not None:
                    startiso = startmat.group(1)
                if stopmat is not None:
                    stopiso = stopmat.group(1)
                if (startiso is None) or (stopiso is None):
                    print(
                        "Could not parse scontrol output for reservation '{}'"
                        .format(reservation)
                    )
                else:
                    from datetime import datetime
                    start = datetime.strptime(startiso, "%Y-%m-%dT%H:%M:%S")
                    stop = datetime.strptime(stopiso, "%Y-%m-%dT%H:%M:%S")
                    now = datetime.now()
                    print(
                        "Reservation '{}' valid from {} to {}".format(
                            reservation,
                            start.isoformat(),
                            stop.isoformat()
                        )
                    )
                    print("Current time is {}".format(now.isoformat()))
                    if (now >= start) and (now < stop):
                        print("Selecting reservation '{}'".format(reservation))
                        nersc_resv = reservation
                    else:
                        print("Reservation '{}' not currently valid".format(reservation))
    else:
        print("Not running at NERSC, slurm jobs disabled.")
    return (nersc_host, nersc_repo, nersc_resv)


def fake_focalplane(
    samplerate=20,
    epsilon=0,
    net=1,
    fmin=0,
    alpha=1,
    fknee=0.05,
    fwhm=30,
    npix=7,
    fov=3.0
):
    """Create a set of fake detectors.

    This generates 7 pixels (14 dets) in a hexagon layout at the boresight
    and with a made up polarization orientation.

    Args:
        None

    Returns:
        (dict):  dictionary of detectors and their properties.

    """
    zaxis = np.array([0, 0, 1.0])
    
    pol_A = hex_pol_angles_qu(npix)
    pol_B = hex_pol_angles_qu(npix, offset=90.0)
    
    dets_A = hex_layout(npix, fov, "", "", pol_A)
    dets_B = hex_layout(npix, fov, "", "", pol_B)
    
    dets = dict()
    for p in range(npix):
        pstr = "{:01d}".format(p)
        for d, layout in zip(["A", "B"], [dets_A, dets_B]):
            props = dict()
            props["quat"] = layout[pstr]["quat"]
            props["epsilon"] = epsilon
            props["rate"] = samplerate
            props["alpha"] = alpha
            props["NET"] = net
            props["fmin"] = fmin
            props["fknee"] = fknee
            props["fwhm_arcmin"] = fwhm
            dname = "{}{}".format(pstr, d)
            dets[dname] = props
    return dets



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import warnings

try:
    import ephem
except:
    ephem = None
import healpy as hp
import numpy as np
from scipy.constants import degree

from .. import qarray as qa
from ..timing import function_timer, Timer
from ..tod import Interval, TOD
from ..healpix import ang2vec
from .pointing_math import quat_equ2ecl, quat_equ2gal, quat_ecl2gal


XAXIS, YAXIS, ZAXIS = np.eye(3)

# FIXME: Once the global package control interface is added,
# move this to that unified location.
tod_buffer_length = 1048576


@function_timer
def simulate_hwp(tod, hwprpm, hwpstep, hwpsteptime):
    """Simulate and cache HWP angle in the TOD"""
    if hwprpm is None and hwpsteptime is None and hwpstep is None:
        # No HWP
        return

    if hwprpm is not None and hwpsteptime is not None:
        raise RuntimeError("choose either continuously rotating or stepped HWP")

    if hwpstep is not None and hwpsteptime is None:
        raise RuntimeError("for a stepped HWP, you must specify the time between steps")

    # compute effective sample rate

    times = tod.local_times()
    dt = np.mean(times[1:-1] - times[0:-2])
    rate = 1.0 / dt
    del times

    if hwprpm is not None:
        # convert to radians / second
        tod._hwprate = hwprpm * 2.0 * np.pi / 60.0
    else:
        tod._hwprate = None

    if hwpstep is not None:
        # convert to radians and seconds
        tod._hwpstep = hwpstep * np.pi / 180.0
        tod._hwpsteptime = hwpsteptime * 60.0
    else:
        tod._hwpstep = None
        tod._hwpsteptime = None

    offset, nsamp = tod.local_samples
    hwp_angle = None
    if tod._hwprate is not None:
        # continuous HWP
        # HWP increment per sample is:
        # (hwprate / samplerate)
        hwpincr = tod._hwprate / rate
        startang = np.fmod(offset * hwpincr, 2 * np.pi)
        hwp_angle = hwpincr * np.arange(nsamp, dtype=np.float64)
        hwp_angle += startang
    elif tod._hwpstep is not None:
        # stepped HWP
        hwp_angle = np.ones(nsamp, dtype=np.float64)
        stepsamples = int(tod._hwpsteptime * rate)
        wholesteps = int(offset / stepsamples)
        remsamples = offset - wholesteps * stepsamples
        curang = np.fmod(wholesteps * tod._hwpstep, 2 * np.pi)
        curoff = 0
        fill = remsamples
        while curoff < nsamp:
            if curoff + fill > nsamp:
                fill = nsamp - curoff
            hwp_angle[curoff:fill] *= curang
            curang += tod._hwpstep
            curoff += fill
            fill = stepsamples

    if hwp_angle is not None:
        # Choose the HWP angle between [0, 2*pi)
        hwp_angle %= 2 * np.pi
        tod.cache.put(tod.HWP_ANGLE_NAME, hwp_angle)

    return


@function_timer
def slew_precession_axis(result, firstsamp=0, samplerate=100.0, degday=1.0):
    """Generate quaternions for constantly slewing precession axis.

    This constructs quaternions which rotates the Z coordinate axis
    to the X/Y plane, and then slowly rotates this.  This can be used
    to generate quaternions for the precession axis used in satellite
    scanning simulations.

    Args:
        result (array): The quaternion array to fill.
        firstsamp (int): The offset in samples from the start
            of rotation.
        samplerate (float): The sampling rate in Hz.
        degday (float): The rotation rate in degrees per day.

    """
    zaxis = np.array([0.0, 0.0, 1.0])

    # this is the increment in radians per sample
    angincr = degday * (np.pi / 180.0) / (24.0 * 3600.0 * samplerate)

    if result.shape[1] != 4:
        raise RuntimeError("Result array has wrong dimensions")

    nsim = result.shape[0]

    # Compute the time-varying quaternions representing the rotation
    # from the coordinate frame to the precession axis frame.  The
    # angle of rotation is fixed (PI/2), but the axis starts at the Y
    # coordinate axis and sweeps.

    buf_off = 0
    buf_n = tod_buffer_length
    while buf_off < nsim:
        if buf_off + buf_n > nsim:
            buf_n = nsim - buf_off
        bslice = slice(buf_off, buf_off + buf_n)

        satang = np.arange(buf_n, dtype=np.float64)
        satang *= angincr
        satang += angincr * (buf_off + firstsamp)
        # satang += angincr * firstsamp + (np.pi / 2)

        cang = np.cos(satang)
        sang = np.sin(satang)

        # this is the time-varying rotation axis
        sataxis = np.concatenate(
            (cang.reshape(-1, 1), sang.reshape(-1, 1), np.zeros((buf_n, 1))), axis=1
        )

        result[bslice, :] = qa.from_vectors(
            np.tile(zaxis, buf_n).reshape(-1, 3), sataxis
        )

        buf_off += buf_n

    return


@function_timer
def satellite_scanning(
    result,
    firstsamp=0,
    samplerate=100.0,
    qprec=None,
    spinperiod=1.0,
    spinangle=85.0,
    precperiod=0.0,
    precangle=0.0,
):
    """Generate boresight quaternions for a generic satellite.

    Given scan strategy parameters and the relevant angles
    and rates, generate an array of quaternions representing
    the rotation of the ecliptic coordinate axes to the
    boresight.

    Args:
        result (array): The quaternion array to fill.
        firstsamp (int): The offset in samples from the start
            of rotation.
        samplerate (float): The sampling rate in Hz.
        qprec (ndarray): If None (the default), then the
            precession axis will be fixed along the
            X axis.  If a 1D array of size 4 is given,
            This will be the fixed quaternion used
            to rotate the Z coordinate axis to the
            precession axis.  If a 2D array of shape
            (nsim, 4) is given, this is the time-varying
            rotation of the Z axis to the precession axis.
        spinperiod (float): The period (in minutes) of the
            rotation about the spin axis.
        spinangle (float): The opening angle (in degrees)
            of the boresight from the spin axis.
        precperiod (float): The period (in minutes) of the
            rotation about the precession axis.
        precangle (float): The opening angle (in degrees)
            of the spin axis from the precession axis.

    """
    if spinperiod > 0.0:
        spinrate = 1.0 / (60.0 * spinperiod)
    else:
        spinrate = 0.0
    spinangle = spinangle * np.pi / 180.0

    if precperiod > 0.0:
        precrate = 1.0 / (60.0 * precperiod)
    else:
        precrate = 0.0
    precangle = precangle * np.pi / 180.0

    xaxis = np.array([1, 0, 0], dtype=np.float64)
    yaxis = np.array([0, 1, 0], dtype=np.float64)
    zaxis = np.array([0, 0, 1], dtype=np.float64)

    if result.shape[1] != 4:
        raise RuntimeError("Result array has wrong dimensions")
    nsim = result.shape[0]

    if qprec is not None:
        if (qprec.shape[0] != nsim) or (qprec.shape[1] != 4):
            raise RuntimeError("qprec array has wrong dimensions")

    buf_off = 0
    buf_n = tod_buffer_length
    while buf_off < nsim:
        if buf_off + buf_n > nsim:
            buf_n = nsim - buf_off
        bslice = slice(buf_off, buf_off + buf_n)

        satrot = np.empty((buf_n, 4), np.float64)
        if qprec is None:
            # in this case, we just have a fixed precession axis, pointing
            # along the ecliptic X axis.
            satrot[:, :] = np.tile(
                qa.rotation(np.array([0.0, 1.0, 0.0]), np.pi / 2), buf_n
            ).reshape(-1, 4)
        elif qprec.flatten().shape[0] == 4:
            # we have a fixed precession axis.
            satrot[:, :] = np.tile(qprec.flatten(), buf_n).reshape(-1, 4)
        else:
            # we have full vector of quaternions
            satrot[:, :] = qprec[bslice, :]

        # Time-varying rotation about precession axis.
        # Increment per sample is
        # (2pi radians) X (precrate) / (samplerate)
        # Construct quaternion from axis / angle form.

        # print("satrot = ", satrot[-1])

        precang = np.arange(buf_n, dtype=np.float64)
        precang += float(buf_off + firstsamp)
        precang *= 2.0 * np.pi * precrate / samplerate
        # print("precang = ", precang[-1])

        cang = np.cos(0.5 * precang)
        sang = np.sin(0.5 * precang)

        precaxis = np.multiply(
            sang.reshape(-1, 1), np.tile(zaxis, buf_n).reshape(-1, 3)
        )

        precrot = np.concatenate((precaxis, cang.reshape(-1, 1)), axis=1)

        # Rotation which performs the precession opening angle
        precopen = qa.rotation(np.array([1.0, 0.0, 0.0]), precangle)

        # Time-varying rotation about spin axis.  Increment
        # per sample is
        # (2pi radians) X (spinrate) / (samplerate)
        # Construct quaternion from axis / angle form.

        spinang = np.arange(buf_n, dtype=np.float64)
        spinang += float(buf_off + firstsamp)
        spinang *= 2.0 * np.pi * spinrate / samplerate

        cang = np.cos(0.5 * spinang)
        sang = np.sin(0.5 * spinang)

        spinaxis = np.multiply(
            sang.reshape(-1, 1), np.tile(zaxis, buf_n).reshape(-1, 3)
        )

        spinrot = np.concatenate((spinaxis, cang.reshape(-1, 1)), axis=1)

        # Rotation which performs the spin axis opening angle

        spinopen = qa.rotation(np.array([1.0, 0.0, 0.0]), spinangle)

        # compose final rotation

        result[bslice, :] = qa.mult(
            satrot, qa.mult(precrot, qa.mult(precopen, qa.mult(spinrot, spinopen)))
        )
        buf_off += buf_n

    return result


class TODHpixSpiral(TOD):
    """Provide a simple generator of fake detector pointing.

    Detector focalplane offsets are specified as a dictionary of 4-element
    ndarrays.  The boresight pointing is a simple looping over HealPix
    ring ordered pixel centers.

    Args:
        mpicomm (mpi4py.MPI.Comm): the MPI communicator over which the
            data are distributed (or None).
        detectors (dictionary): each key is the detector name, and each value
            is a quaternion tuple.
        samples (int):  The total number of samples.
        firsttime (float): starting time of data.
        rate (float): sample rate in Hz.
        nside (int): sky NSIDE to use.
        rot (ndarray):  Extra quaternion to rotate the  quaternions with.
        Other keyword arguments are passed to the parent class constructor.

    """

    def __init__(
        self,
        mpicomm,
        detectors,
        samples,
        firsttime=0.0,
        rate=100.0,
        nside=512,
        rot=None,
        **kwargs
    ):

        self._fp = detectors
        self._detlist = sorted(list(self._fp.keys()))

        props = {"nside": nside}
        super().__init__(mpicomm, self._detlist, samples, meta=props, **kwargs)

        self._firsttime = firsttime
        self._rate = rate
        self._nside = nside
        self._npix = 12 * self._nside * self._nside
        self._rot = rot

    def detoffset(self):
        return {d: np.asarray(self._fp[d]) for d in self._detlist}

    def _get(self, detector, start, n):
        # This class just returns data streams of zeros
        return np.zeros(n, dtype=np.float64)

    def _put(self, detector, start, data, flags):
        raise RuntimeError("cannot write data to simulated data streams")
        return

    def _get_flags(self, detector, start, n):
        return np.zeros(n, dtype=np.uint8)

    def _put_flags(self, detector, start, flags):
        raise RuntimeError("cannot write flags to simulated data streams")
        return

    def _get_common_flags(self, start, n):
        return np.zeros(n, dtype=np.uint8)

    def _put_common_flags(self, start, flags):
        raise RuntimeError("cannot write flags to simulated data streams")
        return

    def _get_times(self, start, n):
        start_abs = self.local_samples[0] + start
        start_time = self._firsttime + float(start_abs) / self._rate
        stop_time = start_time + float(n) / self._rate
        stamps = np.linspace(
            start_time, stop_time, num=n, endpoint=False, dtype=np.float64
        )
        return stamps

    def _put_times(self, start, stamps):
        raise RuntimeError("cannot write timestamps to simulated data streams")
        return

    @function_timer
    def _get_boresight(self, start, n):
        # compute the absolute sample offset
        start_abs = self.local_samples[0] + start

        # pixel offset
        start_pix = int(start_abs % self._npix)
        pixels = np.linspace(start_pix, start_pix + n, num=n, endpoint=False)
        step = (start_abs + pixels) // self._npix
        pixels = np.mod(pixels, self._npix * np.ones(n, dtype=np.int64)).astype(
            np.int64
        )

        # the result of this is normalized
        x, y, z = hp.pix2vec(self._nside, pixels, nest=False)

        # z axis is obviously normalized
        zaxis = np.array([0, 0, 1], dtype=np.float64)
        ztiled = np.tile(zaxis, x.shape[0]).reshape((-1, 3))

        # ... so dir is already normalized
        dir = np.ravel(np.column_stack((x, y, z))).reshape((-1, 3))

        # get the rotation axis
        v = np.cross(ztiled, dir)
        v = v / np.sqrt(np.sum(v * v, axis=1)).reshape((-1, 1))

        # this is the vector-wise dot product
        zdot = np.sum(ztiled * dir, axis=1).reshape((-1, 1))
        ang = 0.5 * np.arccos(zdot)

        # angle element
        s = np.cos(ang)

        # axis
        v *= np.sin(ang)

        # build the normalized quaternion
        boresight = qa.norm(np.concatenate((v, s), axis=1))

        # Add rotation around the boresight
        zrot = qa.rotation(zaxis, np.radians(45) * step)
        boresight = qa.mult(boresight, zrot)
        if self._rot is not None:
            boresight = qa.mult(self._rot, boresight)
        boresight = qa.norm(boresight)

        return boresight

    def _put_boresight(self, start, data):
        raise RuntimeError("cannot write boresight to simulated data streams")
        return

    @function_timer
    def _get_pntg(self, detector, start, n):
        detquat = np.asarray(self._fp[detector])
        boresight = self._get_boresight(start, n)
        data = qa.mult(boresight, detquat)
        return data

    def _put_pntg(self, detector, start, data):
        raise RuntimeError("cannot write data to simulated pointing")
        return

    def _get_position(self, start, n):
        return np.zeros((n, 3), dtype=np.float64)

    def _put_position(self, start, pos):
        raise RuntimeError("cannot write data to simulated position")
        return

    def _get_velocity(self, start, n):
        return np.zeros((n, 3), dtype=np.float64)

    def _put_velocity(self, start, vel):
        raise RuntimeError("cannot write data to simulated velocity")
        return


class TODSatellite(TOD):
    """Provide a simple generator of satellite detector pointing.

    Detector focalplane offsets are specified as a dictionary of 4-element
    ndarrays.  The boresight pointing is a generic 2-angle model.

    Args:
        mpicomm (mpi4py.MPI.Comm): the MPI communicator over which the
            data are distributed (or None).
        detectors (dictionary): each key is the detector name, and each value
            is a quaternion tuple.
        samples (int):  The total number of samples.
        firstsamp (int): The offset in samples from the start
            of the mission.
        firsttime (float): starting time of data.
        rate (float): sample rate in Hz.
        spinperiod (float): The period (in minutes) of the
            rotation about the spin axis.
        spinangle (float): The opening angle (in degrees)
            of the boresight from the spin axis.
        precperiod (float): The period (in minutes) of the
            rotation about the precession axis.
        precangle (float): The opening angle (in degrees)
            of the spin axis from the precession axis.
        coord (str):  Sky coordinate system.  One of
            C (Equatorial), E (Ecliptic) or G (Galactic)
        hwprpm: if None, a constantly rotating HWP is not included.  Otherwise
            it is the rate (in RPM) of constant rotation.
        hwpstep: if None, then a stepped HWP is not included.  Otherwise, this
            is the step in degrees.
        hwpsteptime: The time in minutes between HWP steps.
        All other keyword arguments are passed to the parent constructor.

    """

    def __init__(
        self,
        mpicomm,
        detectors,
        samples,
        firstsamp=0,
        firsttime=0.0,
        rate=100.0,
        spinperiod=1.0,
        spinangle=85.0,
        precperiod=0.0,
        precangle=0.0,
        coord="E",
        hwprpm=None,
        hwpstep=None,
        hwpsteptime=None,
        **kwargs
    ):

        self._fp = detectors
        self._detlist = sorted(list(self._fp.keys()))

        self._firstsamp = firstsamp
        self._firsttime = firsttime
        self._rate = rate
        self._spinperiod = spinperiod
        self._spinangle = spinangle
        self._precperiod = precperiod
        self._precangle = precangle
        self._coord = coord
        self._boresight = None

        props = {
            "spinperiod": spinperiod,
            "spinangle": spinangle,
            "precperiod": precperiod,
            "precangle": precangle,
        }

        # call base class constructor to distribute data
        super().__init__(mpicomm, self._detlist, samples, meta=props, **kwargs)

        self._AU = 149597870.7
        self._radperday = 0.01720209895
        self._radpersec = self._radperday / 86400.0
        self._radinc = self._radpersec / self._rate
        self._earthspeed = self._radpersec * self._AU

        # If HWP parameters are specified, simulate and cache HWP angle

        simulate_hwp(self, hwprpm, hwpstep, hwpsteptime)

    def set_prec_axis(self, qprec=None):
        """Set the fixed or time-varying precession axis.

        This function sets the precession axis for the locally assigned samples.
        It also triggers the generation and caching of the boresight pointing.

        Args:
            qprec (ndarray): If None (the default), then the
                precession axis will be fixed along the
                X axis.  If a 1D array of size 4 is given,
                This will be the fixed quaternion used
                to rotate the Z coordinate axis to the
                precession axis.  If a 2D array of shape
                (local samples, 4) is given, this is the time-varying
                rotation of the Z axis to the precession axis.

        """
        if qprec is not None:
            if (qprec.shape != (4,)) and (qprec.shape != (self.local_samples[1], 4)):
                raise RuntimeError("precession quaternion has incorrect dimensions")

        # generate and cache the boresight pointing
        nsim = self.local_samples[1]
        self._boresight = np.empty(4 * nsim, dtype=np.float64).reshape(-1, 4)
        satellite_scanning(
            self._boresight,
            firstsamp=(self._firstsamp + self.local_samples[0]),
            qprec=qprec,
            samplerate=self._rate,
            spinperiod=self._spinperiod,
            spinangle=self._spinangle,
            precperiod=self._precperiod,
            precangle=self._precangle,
        )

        # Satellite pointing is expressed normally in Ecliptic coordinates.
        # convert coordinate systems in the boresight frame if necessary.
        if self._coord != "E":
            # Add the coordinate system rotation
            if self._coord == "G":
                self._boresight = qa.mult(quat_ecl2gal, self._boresight)
            elif self._coord == "C":
                rot = qa.inv(quat_equ2ecl)
                self._boresight = qa.mult(rot, self._boresight)
            else:
                raise RuntimeError("Unknown coordinate system: {}".format(self._coord))
        return

    def detoffset(self):
        return {d: np.asarray(self._fp[d]) for d in self._detlist}

    def _get_boresight(self, start, n):
        if self._boresight is None:
            raise RuntimeError(
                "you must set the precession axis before reading pointing"
            )
        return self._boresight[start : start + n]

    def _put_boresight(self, start, data):
        raise RuntimeError("cannot write boresight to simulated data streams")
        return

    def _get(self, detector, start, n):
        # This class just returns data streams of zeros
        return np.zeros(n, dtype=np.float64)

    def _put(self, detector, start, data):
        raise RuntimeError("cannot write data to simulated data streams")
        return

    def _get_flags(self, detector, start, n):
        return np.zeros(n, dtype=np.uint8)

    def _put_det_flags(self, detector, start, flags):
        raise RuntimeError("cannot write flags to simulated data streams")
        return

    def _get_common_flags(self, start, n):
        return np.zeros(n, dtype=np.uint8)

    def _put_common_flags(self, start, flags):
        raise RuntimeError("cannot write flags to simulated data streams")
        return

    def _get_hwp_angle(self, start, n):
        if self.cache.exists(self.HWP_ANGLE_NAME):
            angle = self.cache.reference(self.HWP_ANGLE_NAME)[start : start + n]
        else:
            angle = None
        return angle

    def _put_hwp_angle(self, start, flags):
        raise RuntimeError("cannot write HWP angle to simulated data streams")
        return

    def _get_times(self, start, n):
        start_abs = self.local_samples[0] + start
        start_time = self._firsttime + float(start_abs) / self._rate
        stop_time = start_time + float(n) / self._rate
        stamps = np.linspace(
            start_time, stop_time, num=n, endpoint=False, dtype=np.float64
        )
        return stamps

    def _put_times(self, start, stamps):
        raise RuntimeError("cannot write timestamps to simulated data streams")
        return

    def _get_pntg(self, detector, start, n):
        boresight = self._get_boresight(start, n)
        detquat = self._fp[detector]
        data = qa.mult(boresight, detquat)
        return data

    def _put_pntg(self, detector, start, data):
        raise RuntimeError("cannot write data to simulated pointing")
        return

    def _get_position(self, start, n):
        # For this simple class, assume that the Earth is located
        # along the X axis at time == 0.0s.  We also just use the
        # mean values for distance and angular speed.  Classes for
        # real experiments should obviously use ephemeris data.
        rad = np.fmod((start - self._firsttime) * self._radpersec, 2.0 * np.pi)
        ang = self._radinc * np.arange(n, dtype=np.float64) + rad
        x = self._AU * np.cos(ang)
        y = self._AU * np.sin(ang)
        z = np.zeros_like(x)
        return np.ravel(np.column_stack((x, y, z))).reshape((-1, 3))

    def _put_position(self, start, pos):
        raise RuntimeError("cannot write data to simulated position")
        return

    def _get_velocity(self, start, n):
        # For this simple class, assume that the Earth is located
        # along the X axis at time == 0.0s.  We also just use the
        # mean values for distance and angular speed.  Classes for
        # real experiments should obviously use ephemeris data.
        rad = np.fmod((start - self._firsttime) * self._radpersec, 2.0 * np.pi)
        ang = self._radinc * np.arange(n, dtype=np.float64) + rad + (0.5 * np.pi)
        x = self._earthspeed * np.cos(ang)
        y = self._earthspeed * np.sin(ang)
        z = np.zeros_like(x)
        return np.ravel(np.column_stack((x, y, z))).reshape((-1, 3))

    def _put_velocity(self, start, vel):
        raise RuntimeError("cannot write data to simulated velocity")
        return


class TODGround(TOD):
    """Provide a simple generator of ground-based detector pointing.

    Detector focalplane offsets are specified as a dictionary of
    4-element ndarrays.  The boresight pointing is a generic
    2-angle model.

    Args:
        mpicomm (mpi4py.MPI.Comm): the MPI communicator over which the
            data is distributed (or None).
        detectors (dictionary): each key is the detector name, and each
            value is a quaternion tuple.
        samples (int):  The total number of samples.
        boresight_angle (float):  Extra rotation to apply around the
             boresight [degrees]
        firsttime (float): starting time of data.
        rate (float): sample rate in Hz.
        site_lon (float/str): Observing site Earth longitude in radians
            or a pyEphem string.
        site_lat (float/str): Observing site Earth latitude in radians
            or a pyEphem string.
        site_alt (float/str): Observing site Earth altitude in meters.
        azmin (float):  Starting azimuth of the constant elevation scan
            or el-nod
        azmax (float): Ending azimuth of the constant elevation scan.
            If `None`, no CES is performed.
        el (float):  Elevation of the scan.
        el_nod (string or list of float): List of elevations to acquire
            relative to `el`.  First and last elevation will always be
            `el`.  Beginning el-nod is always performed at `azmin`.
            Ending el-nod is always performed wherever the science scan
            ends.  [degrees]
        start_with_elnod (bool) : Perform el-nod before the scan
        end_with_elnod (bool) : Perform el-nod after the scan
        el_mod_step (float) : If non-zero, scanning elevation will be
            stepped after each scan pair (upon returning to starting
            azimuth). [degrees]
        el_mod_rate (float) : If non-zero, observing elevation will be
            continuously modulated during the science scan. [Hz]
        el_mod_amplitude (float) : Range of elevation modulation when
            `el_mod_rate` is non-zero. [degrees]
        el_mod_sine (bool) : Elevation modulation should assume a
            sine-wave shape.  If False, use a triangular wave.
        scanrate (float): Sky scanning rate in degrees / second.
        scanrate_el (float): Sky elevation scanning rate in
            degrees / second.  If None, will default to `scanrate`.
        scan_accel (float): Mount scanning rate acceleration in
            degrees / second^2 for the turnarounds.
        scan_accel_el (float): Mount elevation rate acceleration in
            degrees / second^2.  If None, will default to `scan_accel`.
        cosecant_modulation (bool): Modulate the scan rate according to
             1/sin(az) to achieve uniform integration depth.
        CES_start (float): Start time of the constant elevation scan
        CES_stop (float): Stop time of the constant elevation scan
        sun_angle_min (float): Minimum angular distance for the scan and
            the Sun [degrees].
        sampsizes (list):  Optional list of sample chunk sizes which
            cannot be split.
        sampbreaks (list):  Optional list of hard breaks in the sample
            distribution.
        coord (str):  Sky coordinate system.  One of
            C (Equatorial), E (Ecliptic) or G (Galactic)
        report_timing (bool):  Report the time spent simulating the scan
            and translating the pointing.
        hwprpm (float): If None, a constantly rotating HWP is not
            included. Otherwise it is the rate (in RPM) of constant
            rotation.
        hwpstep (float): If None, then a stepped HWP is not included.
            Otherwise, this is the step in degrees.
        hwpsteptime (float): The time in minutes between HWP steps.
        All other keyword arguments are passed to the parent constructor.

    """

    TURNAROUND = 1
    LEFTRIGHT_SCAN = 2
    RIGHTLEFT_SCAN = 4
    LEFTRIGHT_TURNAROUND = LEFTRIGHT_SCAN + TURNAROUND
    RIGHTLEFT_TURNAROUND = RIGHTLEFT_SCAN + TURNAROUND
    SUN_UP = 8
    SUN_CLOSE = 16
    ELNOD = 32

    @function_timer
    def __init__(
        self,
        mpicomm,
        detectors,
        samples,
        boresight_angle=0,
        firsttime=0.0,
        rate=100.0,
        site_lon=0,
        site_lat=0,
        site_alt=0,
        el=None,
        azmin=None,
        azmax=None,
        el_nod=None,
        start_with_elnod=True,
        end_with_elnod=False,
        el_mod_step=0,
        el_mod_rate=0,
        el_mod_amplitude=1,
        el_mod_sine=False,
        scanrate=1,
        scanrate_el=None,
        scan_accel=0.1,
        scan_accel_el=None,
        CES_start=None,
        CES_stop=None,
        el_min=0,
        sun_angle_min=90,
        sampsizes=None,
        sampbreaks=None,
        coord="C",
        report_timing=True,
        hwprpm=None,
        hwpstep=None,
        hwpsteptime=None,
        cosecant_modulation=False,
        **kwargs
    ):
        if samples < 1:
            raise RuntimeError(
                "TODGround must be instantiated with a positive number of "
                "samples, not samples == {}".format(samples)
            )

        if ephem is None:
            raise RuntimeError("Cannot instantiate a TODGround object without pyephem.")

        if sampsizes is not None or sampbreaks is not None:
            raise RuntimeError(
                "TODGround will synthesize the sizes to match the subscans."
            )

        testvec = np.array([el, azmin, azmax])
        if azmax:
            if not np.all(testvec):
                raise RuntimeError(
                    "Simulating a CES requires `el`, `azmin` and `azmax`"
                )
            do_ces = True
        else:
            do_ces = False

        if el_nod is not None:
            if azmin is None:
                raise RuntimeError("Simulating an el-nod requires `azmin`")
            do_elnod = True
        else:
            do_elnod = False

        if scanrate_el is None:
            scanrate_el = scanrate
        if scan_accel_el is None:
            scan_accel_el = scan_accel

        if not do_ces and not do_elnod:
            raise RuntimeError("You must provide parameters for a CES and/or el-nod")

        if CES_start is not None:
            warnings.warn("User should not set CES_start", DeprecationWarning)
            CES_start = None
        if CES_stop is not None:
            warnings.warn("User should not set CES_stop", DeprecationWarning)
            CES_stop = None

        self._boresight_angle = boresight_angle * degree
        self._firsttime = firsttime
        self._rate = rate
        self._site_lon = site_lon
        self._site_lat = site_lat
        self._site_alt = site_alt
        if azmax is not None and samples > 0:
            if el < 1 or el > 89:
                raise RuntimeError("Impossible CES at {:.2f} degrees".format(el))
            if azmin == azmax:
                raise RuntimeError("TODGround CES requires non-empty azimuth range")
            self._azmin_ces = azmin * degree
            self._azmax_ces = azmax * degree
            self._el_ces = el * degree
        else:
            self._el_ces = None
        if do_elnod:
            if not start_with_elnod and not end_with_elnod:
                raise RuntimeError("El-not requested without start or end")
            try:
                # Try parsing as a comma-separated string
                el_nod_list = [float(x) for x in el_nod.split(",")]
            except:
                # Must already be an iterable
                el_nod_list = list(el_nod)
            el_nod = np.array(el_nod_list) + el
            if np.any(el_nod < 1) or np.any(el_nod > 90):
                raise RuntimeError("Impossible el-nod at {} degrees".format(el_nod))
            el_nod = list(el_nod)
            if el is not None:
                if abs(el_nod[0] - el) > 1e-3:
                    el_nod.insert(0, el)
                if abs(el_nod[-1] - el) > 1e-3:
                    el_nod.append(el)
                el_nod = np.array(el_nod)
            self._elnod_el = el_nod * degree
            self._elnod_az = np.zeros_like(el_nod) + azmin * degree
        else:
            self._elnod_el = None
            self._elnod_az = None
        self._el_mod_step = el_mod_step * degree
        self._el_mod_rate = el_mod_rate
        self._el_mod_amplitude = el_mod_amplitude * degree
        self._el_mod_sine = el_mod_sine
        self._scanrate = scanrate * degree
        self._scan_accel = scan_accel * degree
        self._scanrate_el = scanrate_el * degree
        self._scan_accel_el = scan_accel_el * degree
        self._CES_start = CES_start
        self._CES_stop = CES_stop
        self._sun_angle_min = sun_angle_min
        if coord not in "CEG":
            raise RuntimeError("Unknown coordinate system: {}".format(coord))
        self._coord = coord
        self._report_timing = report_timing
        self._cosecant_modulation = cosecant_modulation

        self._observer = ephem.Observer()
        self._observer.lon = self._site_lon
        self._observer.lat = self._site_lat
        self._observer.elevation = self._site_alt  # In meters
        self._observer.epoch = ephem.J2000  # "2000"
        # self._observer.epoch = -9786 # EOD
        self._observer.compute_pressure()

        self._min_az = None
        self._max_az = None
        self._min_el = None
        self._min_el = None

        self._boresight_azel = None
        self._boresight = None

        # Set the boresight pointing based on the given scan parameters

        timer = Timer()
        if self._report_timing:
            if mpicomm is not None:
                mpicomm.Barrier()
            timer.start()

        self._times = np.array([])
        self._commonflags = np.array([], dtype=np.uint8)
        self._az = np.array([])
        self._el = np.array([])

        nsample_elnod = 0
        if start_with_elnod:
            # Begin with an el-nod
            nsample_elnod = self.simulate_elnod(
                self._firsttime, azmin * degree, el * degree
            )
            if nsample_elnod > 0:
                t_elnod = self._times[-1] - self._times[0]
                # Shift the time stamps so that the CES starts at the prescribed time
                self._times -= t_elnod
                self._firsttime -= t_elnod

        nsample_ces = self.simulate_scan(samples)

        if end_with_elnod and self._elnod_az is not None:
            # Append en el-nod after the CES
            self._elnod_az[:] = self._az[-1]
            nsample_elnod = self.simulate_elnod(
                self._times[-1], self._az[-1], self._el[-1]
            )
        self._lasttime = self._times[-1]
        samples = self._times.size

        if self._report_timing:
            if mpicomm is not None:
                mpicomm.Barrier()
            if mpicomm is None or mpicomm.rank == 0:
                timer.report_clear("TODGround: simulate scan")

        # Create a list of subscans that excludes the turnarounds.
        # All processes in the group still have all samples.

        self.subscans = []
        self._subscan_min_length = 10  # in samples
        for istart, istop in zip(self._stable_starts, self._stable_stops):
            if istop - istart < self._subscan_min_length or istart < nsample_elnod:
                self._commonflags[istart:istop] |= self.TURNAROUND
                continue
            start = self._firsttime + istart / self._rate
            stop = self._firsttime + istop / self._rate
            self.subscans.append(
                Interval(start=start, stop=stop, first=istart, last=istop - 1)
            )

        if len(self._stable_stops) > 0:
            self._commonflags[self._stable_stops[-1] :] |= self.TURNAROUND

        if np.sum((self._commonflags & self.TURNAROUND) == 0) == 0 and do_ces:
            raise RuntimeError(
                "The entire TOD is flagged as turnaround. Samplerate too low "
                "({} Hz) or scanrate too high ({} deg/s)?"
                "".format(rate, scanrate)
            )

        if self._report_timing:
            if mpicomm is not None:
                mpicomm.Barrier()
            if mpicomm is None or mpicomm.rank == 0:
                timer.report_clear("TODGround: list valid intervals")

        self._fp = detectors
        self._detlist = sorted(list(self._fp.keys()))

        # call base class constructor to distribute data

        props = {
            "site_lon": site_lon,
            "site_lat": site_lat,
            "site_alt": site_alt,
            "azmin": azmin,
            "azmax": azmax,
            "el": el,
            "scanrate": scanrate,
            "scan_accel": scan_accel,
            "el_min": el_min,
            "sun_angle_min": sun_angle_min,
        }
        super().__init__(
            mpicomm,
            self._detlist,
            samples,
            sampsizes=[samples],
            sampbreaks=None,
            meta=props,
            **kwargs
        )

        self._AU = 149597870.7
        self._radperday = 0.01720209895
        self._radpersec = self._radperday / 86400.0
        self._radinc = self._radpersec / self._rate
        self._earthspeed = self._radpersec * self._AU

        if self._report_timing:
            if mpicomm is not None:
                mpicomm.Barrier()
            if mpicomm is None or mpicomm.rank == 0:
                timer.report_clear("TODGround: call base class constructor")

        self.translate_pointing()

        self.crop_vectors()

        if self._report_timing:
            if mpicomm is not None:
                mpicomm.Barrier()
            if mpicomm is None or mpicomm.rank == 0:
                timer.report_clear("TODGround: translate scan pointing")

        # If HWP parameters are specified, simulate and cache HWP angle

        simulate_hwp(self, hwprpm, hwpstep, hwpsteptime)

        return

    @function_timer
    def scan_time(self, coord_in, coord_out, scanrate, scan_accel):
        """Given a coordinate range and scan parameters, determine
        the time taken to scan between them, assuming that the scan
        begins and ends at rest.
        """
        d = np.abs(coord_in - coord_out)
        t_accel = scanrate / scan_accel
        d_accel = 0.5 * scan_accel * t_accel ** 2
        if 2 * d_accel > d:
            # No time to reach scan speed
            d_accel = d / 2
            t_accel = np.sqrt(2 * d_accel / scan_accel)
            t_coast = 0
        else:
            d_coast = d - 2 * d_accel
            t_coast = d_coast / scanrate
        return 2 * t_accel + t_coast

    @function_timer
    def scan_profile(
        self, coord_in, coord_out, scanrate, scan_accel, times, nstep=10000
    ):
        """scan between the coordinates assuming that the scan begins
        and ends at rest.  If there is more time than is needed, wait at the end.
        """
        if np.abs(coord_in - coord_out) < 1e-6:
            return np.zeros(times.size) + coord_out

        d = np.abs(coord_in - coord_out)
        t_accel = scanrate / scan_accel
        d_accel = 0.5 * scan_accel * t_accel ** 2
        if 2 * d_accel > d:
            # No time to reach scan speed
            d_accel = d / 2
            t_accel = np.sqrt(2 * d_accel / scan_accel)
            t_coast = 0
            # Scan rate at the end of acceleration
            scanrate = t_accel * scan_accel
        else:
            d_coast = d - 2 * d_accel
            t_coast = d_coast / scanrate
        if coord_in > coord_out:
            scanrate *= -1
            scan_accel *= -1
        #
        t = []
        coord = []
        # Acceleration
        t.append(np.linspace(times[0], times[0] + t_accel, nstep))
        coord.append(coord_in + 0.5 * scan_accel * (t[-1] - t[-1][0]) ** 2)
        # Coasting
        if t_coast > 0:
            t.append(np.linspace(t[-1][-1], t[-1][-1] + t_coast, 3))
            coord.append(coord[-1][-1] + scanrate * (t[-1] - t[-1][0]))
        # Deceleration
        t.append(np.linspace(t[-1][-1], t[-1][-1] + t_accel, nstep))
        coord.append(
            coord[-1][-1]
            + scanrate * (t[-1] - t[-1][0])
            - 0.5 * scan_accel * (t[-1] - t[-1][0]) ** 2
        )
        # Wait
        if t[-1][-1] < times[-1]:
            t.append(np.linspace(t[-1][-1], times[-1], 3))
            coord.append(np.zeros(3) + coord_out)

        # Interpolate to the given time stamps
        t = np.hstack(t)
        coord = np.hstack(coord)

        return np.interp(times, t, coord)

    @function_timer
    def scan_between(self, time_start, azel_in, azel_out, nstep=10000):
        """Using self._scanrate, self._scan_accel, self._scanrate_el
        and self._scan_accel_el, simulate motion between the two
        coordinates
        """

        az1, el1 = azel_in
        az2, el2 = azel_out
        az_time = self.scan_time(az1, az2, self._scanrate, self._scan_accel)
        el_time = self.scan_time(el1, el2, self._scanrate_el, self._scan_accel_el)
        time_tot = max(az_time, el_time)
        times = np.linspace(0, time_tot, nstep)
        az = self.scan_profile(az1, az2, self._scanrate, self._scan_accel, times)
        el = self.scan_profile(el1, el2, self._scanrate_el, self._scan_accel_el, times)

        return times + time_start, az, el

    @function_timer
    def __del__(self):
        try:
            del self._boresight_azel
        except:
            pass
        try:
            del self._boresight
        except:
            pass
        try:
            del self._az
        except:
            pass
        try:
            del self._commonflags
        except:
            pass
        try:
            self.cache.clear()
        except:
            pass

    def to_JD(self, t):
        """
        Convert TOAST UTC time stamp to Julian date
        """
        return t / 86400.0 + 2440587.5

    def to_DJD(self, t):
        """
        Convert TOAST UTC time stamp to Dublin Julian date used
        by pyEphem.
        """
        return self.to_JD(t) - 2415020

    @property
    def scan_range(self):
        """
        (tuple):  The extent of the boresight pointing as (min_az, max_az,
            min_el, max_el) in radians.  Includes turnarounds.
        """
        return self._min_az, self._max_az, self._min_el, self._max_el

    @function_timer
    def simulate_elnod(self, t_start, az_start, el_start):
        """Simulate an el-nod, if one was requested"""

        if self._elnod_el is not None:
            time_last = t_start
            az_last = az_start
            el_last = el_start
            t = []
            az = []
            el = []
            for az_new, el_new in zip(self._elnod_az, self._elnod_el):
                if np.abs(az_last - az_new) > 1e-3 or np.abs(el_last - el_new) > 1e-3:
                    tvec, azvec, elvec = self.scan_between(
                        time_last, (az_last, el_last), (az_new, el_new)
                    )
                    t.append(tvec)
                    az.append(azvec)
                    el.append(elvec)
                    time_last = tvec[-1]
                az_last = az_new
                el_last = el_new

            t = np.hstack(t)
            az = np.hstack(az)
            el = np.hstack(el)

            # Store the scan range.  We use the high resolution elevation
            # so actual sampling rate will not change the range.
            self.update_scan_range(az, el)

            # Sample t/az/el down to the sampling rate
            nsample_elnod = int((t[-1] - t[0]) * self._rate)
            t_sample = np.arange(nsample_elnod) / self._rate + t_start
            az_sample = np.interp(t_sample, t, az)
            el_sample = np.interp(t_sample, t, el)
            commonflags_sample = np.zeros(nsample_elnod, dtype=np.uint8) + (
                self.ELNOD | self.TURNAROUND
            )
            self._times = np.hstack([self._times, t_sample])
            self._commonflags = np.hstack([self._commonflags, commonflags_sample])
            self._az = np.hstack([self._az, az_sample])
            self._el = np.hstack([self._el, el_sample])
        else:
            nsample_elnod = 0

        return nsample_elnod

    @function_timer
    def update_scan_range(self, az, el):
        if self._min_az is None:
            self._min_az = np.amin(az)
            self._max_az = np.amax(az)
            self._min_el = np.amin(el)
            self._max_el = np.amax(el)
        else:
            self._min_az = min(self._min_az, np.amin(az))
            self._max_az = max(self._max_az, np.amax(az))
            self._min_el = min(self._min_el, np.amin(el))
            self._max_el = max(self._max_el, np.amax(el))
        return

    @function_timer
    def oscillate_el(self, times, az, el):
        """ Simulate oscillating elevation """

        tt = times - times[0]
        if self._el_mod_sine:
            # elevation is modulated along a sine wave
            angular_rate = 2 * np.pi * self._el_mod_rate
            el += self._el_mod_amplitude * np.sin(tt * angular_rate)

            # Check that we did not breach tolerances
            el_rate_max = np.amax(np.abs(np.diff(el)) / np.diff(tt))
            if np.any(el_rate_max > self._scanrate_el):
                raise RuntimeError(
                    "Elevation oscillation requires {:.2f} deg/s but "
                    "mount only allows {:.2f} deg/s".format(
                        np.degrees(el_rate_max), np.degrees(self._scanrate_el)
                    )
                )
        else:
            # elevation is modulated using a constant rate.  We need to
            # calculate the appropriate scan rate to achive the desired
            # amplitude and period
            t_mod = 1 / self._el_mod_rate
            # determine scan rate needed to reach given amplitude in given time
            a = self._scan_accel_el
            b = -0.5 * self._scan_accel_el * t_mod
            c = 2 * self._el_mod_amplitude
            if b ** 2 - 4 * a * c < 0:
                raise RuntimeError(
                    "Cannot perform {:.2f} deg elevation oscillation in {:.2f} s "
                    "with {:.2f} deg/s^2 acceleration".format(
                        np.degrees(self._el_mod_amplitude * 2),
                        np.degrees(self._scan_accel_el),
                    )
                )
            root1 = (-b - np.sqrt(b ** 2 - 4 * a * c)) / (2 * a)
            root2 = (-b + np.sqrt(b ** 2 - 4 * a * c)) / (2 * a)
            if root1 > 0:
                t_accel = root1
            else:
                t_accel = root2
            t_scan = 0.5 * t_mod - 2 * t_accel
            scanrate = t_accel * self._scan_accel_el
            if scanrate > self._scanrate_el:
                raise RuntimeError(
                    "Elevation oscillation requires {:.2f} > {:.2f} deg/s "
                    "scan rate".format(
                        np.degrees(scanrate),
                        np.degrees(self._scanrate_el),
                    )
                )

            # simulate a high resolution scan to interpolate

            t_interp = []
            el_interp = []
            n = 1000
            # accelerate
            t = np.linspace(0, t_accel, n)
            t_interp.append(t)
            el_interp.append(0.5 * self._scan_accel_el * t ** 2)
            # scan
            t_last = t_interp[-1][-1]
            el_last = el_interp[-1][-1]
            t = np.linspace(0, t_scan, 2)
            t_interp.append(t_last + t)
            el_interp.append(el_last + t * scanrate)
            # decelerate
            t_last = t_interp[-1][-1]
            el_last = el_interp[-1][-1]
            t = np.linspace(0, 2 * t_accel, n)
            t_interp.append(t_last + t)
            el_interp.append(
                el_last + scanrate * t - 0.5 * self._scan_accel_el * t ** 2
            )
            # scan
            t_last = t_interp[-1][-1]
            el_last = el_interp[-1][-1]
            t = np.linspace(0, t_scan, 2)
            t_interp.append(t_last + t)
            el_interp.append(el_last - t * scanrate)
            # decelerate
            t_last = t_interp[-1][-1]
            el_last = el_interp[-1][-1]
            t = np.linspace(0, t_accel, n)
            t_interp.append(t_last + t)
            el_interp.append(
                el_last - scanrate * t + 0.5 * self._scan_accel_el * t ** 2
            )

            t_interp = np.hstack(t_interp)
            el_interp = np.hstack(el_interp)

            # finally, modulate the elevation around the input vector
            # start time is set to middle of the first stable scan
            # *** without a proper accelerating phase ***
            tt += t_accel + 0.5 * t_scan
            el += np.interp(tt % t_mod, t_interp, el_interp) - self._el_mod_amplitude

        return

    @function_timer
    def step_el(self, times, az, el):
        """ Simulate elevation steps after each scan pair """

        sign = np.sign(self._el_mod_step)
        el_step = np.abs(self._el_mod_step)

        # simulate a single elevation step at high resolution

        t_accel = self._scanrate_el / self._scan_accel_el
        el_accel = 0.5 * self._scan_accel_el * t_accel ** 2
        if el_step > 2 * el_accel:
            # Step is large enough to reach elevation scan rate
            t_scan = (el_step - 2 * el_accel) / self._scanrate_el
        else:
            # Only partial acceleration and deceleration
            el_accel = np.abs(self._el_mod_step) / 2
            t_accel = np.sqrt(2 * el_accel / self._scan_accel_el)
            t_scan = 0

        t_interp = []
        el_interp = []
        n = 1000
        # accelerate
        t = np.linspace(0, t_accel, n)
        t_interp.append(t)
        el_interp.append(0.5 * self._scan_accel_el * t ** 2)
        # scan
        if t_scan > 0:
            t_last = t_interp[-1][-1]
            el_last = el_interp[-1][-1]
            t = np.linspace(0, t_scan, n)
            t_interp.append(t_last + t)
            el_interp.append(el_last + t * self._scanrate_el)
        # decelerate
        t_last = t_interp[-1][-1]
        el_last = el_interp[-1][-1]
        el_rate_last = self._scan_accel_el * t_accel
        t = np.linspace(0, t_accel, n)
        t_interp.append(t_last + t)
        el_interp.append(
            el_last + el_rate_last * t - 0.5 * self._scan_accel_el * t ** 2
        )

        t_interp = np.hstack(t_interp)
        t_interp -= t_interp[t_interp.size // 2]
        el_interp = sign * np.hstack(el_interp)

        # isolate steps

        daz = np.diff(az)
        ind = np.where(daz[1:] * daz[:-1] < 0)[0] + 1
        ind = ind[1::2]

        # Modulate the elevation at each step

        for istep in ind:
            tstep = times[istep]
            el += np.interp(times - tstep, t_interp, el_interp)

        return

    @function_timer
    def simulate_scan(self, samples):
        """Simulate el-nod and/or a constant elevation scan, either constant rate or
        1/sin(az)-modulated.

        """

        if self._el_ces is None:
            # User only wanted an el-nod
            self._stable_starts = []
            self._stable_stops = []
            return 0

        if samples <= 0:
            raise RuntimeError("CES requires a positive number of samples")

        if len(self._times) == 0:
            self._CES_start = self._firsttime
        else:
            self._CES_start = self._times[-1] + 1 / self._rate

        # Begin by simulating one full scan with turnarounds.
        # It will be used to interpolate the full CES.
        # `nstep` is the number of steps used for one sweep or one
        # turnaround.

        nstep = 10000

        azmin, azmax = [self._azmin_ces, self._azmax_ces]
        if self._cosecant_modulation:
            # We always simulate a rising cosecant scan and then
            # mirror it if necessary
            azmin %= np.pi
            azmax %= np.pi
            if azmin > azmax:
                raise RuntimeError(
                    "Cannot scan across zero meridian with cosecant-modulated scan"
                )
        elif azmax < azmin:
            azmax += 2 * np.pi
        t = self._CES_start
        all_t = []
        all_az = []
        all_flags = []
        # translate scan rate from sky to mount coordinates
        base_rate = self._scanrate / np.cos(self._el_ces)
        # scan acceleration is already in the mount coordinates
        scan_accel = self._scan_accel

        # left-to-right

        tvec = []
        azvec = []
        t0 = t
        if self._cosecant_modulation:
            t1 = t0 + (np.cos(azmin) - np.cos(azmax)) / base_rate
            tvec = np.linspace(t0, t1, nstep, endpoint=True)
            azvec = np.arccos(np.cos(azmin) + base_rate * t0 - base_rate * tvec)
        else:
            # Constant scanning rate, only requires two data points
            t1 = t0 + (azmax - azmin) / base_rate
            tvec = np.array([t0, t1])
            azvec = np.array([azmin, azmax])
        all_t.append(np.array(tvec))
        all_az.append(np.array(azvec))
        all_flags.append(np.zeros(tvec.size, dtype=np.uint8) | self.LEFTRIGHT_SCAN)
        t = t1

        # turnaround

        t0 = t
        if self._cosecant_modulation:
            dazdt = base_rate / np.abs(np.sin(azmax))
        else:
            dazdt = base_rate
        t1 = t0 + 2 * dazdt / scan_accel
        tvec = np.linspace(t0, t1, nstep, endpoint=True)[1:]
        azvec = azmax + (tvec - t0) * dazdt - 0.5 * scan_accel * (tvec - t0) ** 2
        all_t.append(tvec[:-1])
        all_az.append(azvec[:-1])
        all_flags.append(
            np.zeros(tvec.size - 1, dtype=np.uint8) | self.LEFTRIGHT_TURNAROUND
        )
        t = t1

        # right-to-left

        tvec = []
        azvec = []
        t0 = t
        if self._cosecant_modulation:
            t1 = t0 + (np.cos(azmin) - np.cos(azmax)) / base_rate
            tvec = np.linspace(t0, t1, nstep, endpoint=True)
            azvec = np.arccos(np.cos(azmax) - base_rate * t0 + base_rate * tvec)
        else:
            # Constant scanning rate, only requires two data points
            t1 = t0 + (azmax - azmin) / base_rate
            tvec = np.array([t0, t1])
            azvec = np.array([azmax, azmin])
        all_t.append(np.array(tvec))
        all_az.append(np.array(azvec))
        all_flags.append(np.zeros(tvec.size, dtype=np.uint8) | self.RIGHTLEFT_SCAN)
        t = t1

        # turnaround

        t0 = t
        if self._cosecant_modulation:
            dazdt = base_rate / np.abs(np.sin(azmin))
        else:
            dazdt = base_rate
        t1 = t0 + 2 * dazdt / scan_accel
        tvec = np.linspace(t0, t1, nstep, endpoint=True)[1:]
        azvec = azmin - (tvec - t0) * dazdt + 0.5 * scan_accel * (tvec - t0) ** 2
        all_t.append(tvec)
        all_az.append(azvec)
        all_flags.append(
            np.zeros(tvec.size, dtype=np.uint8) | self.RIGHTLEFT_TURNAROUND
        )

        # Concatenate

        tvec = np.hstack(all_t)
        azvec = np.hstack(all_az)
        flags = np.hstack(all_flags)

        # Limit azimuth to [-2pi, 2pi] but do not
        # introduce discontinuities with modulo.

        if np.amin(azvec) < -2 * np.pi:
            azvec += 2 * np.pi
        if np.amax(azvec) > 2 * np.pi:
            azvec -= 2 * np.pi

        if self._cosecant_modulation and self._azmin_ces > np.pi:
            # We always simulate a rising cosecant scan and then
            # mirror it if necessary
            azvec += np.pi

        # Store the scan range.  We use the high resolution azimuth so the
        # actual sampling rate will not change the range.

        self.update_scan_range(azvec, self._el_ces)

        # Now interpolate the simulated scan to timestamps

        times = self._CES_start + np.arange(samples) / self._rate
        tmin, tmax = tvec[0], tvec[-1]
        tdelta = tmax - tmin
        az_sample = np.interp((times - tmin) % tdelta, tvec - tmin, azvec)
        if self._cosecant_modulation and self._azmin_ces > np.pi:
            # We always simulate a rising cosecant scan and then
            # mirror it if necessary
            az_sample += np.pi

        el_sample = np.zeros_like(az_sample) + self._el_ces
        if self._el_mod_rate != 0:
            self.oscillate_el(times, az_sample, el_sample)
        if self._el_mod_step != 0:
            self.step_el(times, az_sample, el_sample)

        offset = self._times.size
        self._times = np.hstack([self._times, times])
        self._az = np.hstack([self._az, az_sample])
        self._el = np.hstack([self._el, el_sample])
        ind = np.searchsorted(tvec - tmin, (times - tmin) % tdelta)
        ind[ind == tvec.size] = tvec.size - 1
        self._commonflags = np.hstack([self._commonflags, flags[ind]]).astype(np.uint8)

        # Subscan start indices

        turnflags = flags[ind] & self.TURNAROUND
        self._stable_starts = (
            np.argwhere(np.logical_and(turnflags[:-1] != 0, turnflags[1:] == 0)).ravel()
            + 1
        )
        if turnflags[0] == 0:
            self._stable_starts = np.hstack([[0], self._stable_starts])
        self._stable_stops = (
            np.argwhere(np.logical_and(turnflags[:-1] == 0, turnflags[1:] != 0)).ravel()
            + 2
        )
        if turnflags[-1] == 0:
            self._stable_stops = np.hstack([self._stable_stops, [samples]])
        self._stable_starts += offset
        self._stable_stops += offset

        self._CES_stop = self._times[-1]

        return samples

    @function_timer
    def translate_pointing(self):
        """Translate Az/El into bore sight quaternions

        Translate the azimuth and elevation into bore sight quaternions.

        """
        # At this point, all processes still have all of the scan
        nsamp = len(self._az)
        rank = 0
        ntask = 1
        if self._mpicomm is not None:
            rank = self._mpicomm.rank
            ntask = self._mpicomm.size
        nsamp_task = nsamp // ntask + 1
        my_start = rank * nsamp_task
        my_stop = min(my_start + nsamp_task, nsamp)
        my_nsamp = max(0, my_stop - my_start)
        my_ind = slice(my_start, my_stop)

        # Remember that the azimuth is measured clockwise and the
        # longitude counter-clockwise
        my_azelquats = qa.from_angles(
            np.pi / 2 - self._el[my_ind],
            -(self._az[my_ind]),
            np.zeros(my_nsamp),
            IAU=False,
        )

        if self._boresight_angle != 0:
            zaxis = np.array([0, 0, 1.0])
            rot = qa.rotation(zaxis, self._boresight_angle)
            my_azelquats = qa.mult(my_azelquats, rot)

        azelquats = None
        if self._mpicomm is None:
            azelquats = my_azelquats
        else:
            azelquats = np.vstack(self._mpicomm.allgather(my_azelquats))
        self._boresight_azel = azelquats

        my_times = self.local_times()[my_ind]
        azel2radec_times, azel2radec_quats = self._get_azel2radec_quats()
        my_azel2radec_quats = qa.slerp(my_times, azel2radec_times, azel2radec_quats)
        my_quats = qa.mult(my_azel2radec_quats, my_azelquats)
        del my_azelquats

        quats = None
        if self._mpicomm is None:
            quats = my_quats
        else:
            quats = np.vstack(self._mpicomm.allgather(my_quats))

        self._boresight = quats
        del my_quats
        return

    @function_timer
    def crop_vectors(self):
        """Crop the TOD vectors.

        Crop the TOD vectors to match the sample range assigned to this task.

        """
        offset, n = self.local_samples
        ind = slice(offset, offset + n)

        self._times = self.cache.put("times", self._times[ind], replace=True)
        self._az = self.cache.put("az", self._az[ind])
        self._el = self.cache.put("el", self._el[ind])
        self._commonflags = self.cache.put(
            "common_flags", self._commonflags[ind], replace=True
        )
        self._boresight_azel = self.cache.put(
            "boresight_azel", self._boresight_azel[ind]
        )
        self._boresight = self.cache.put("boresight_radec", self._boresight[ind])
        return

    @function_timer
    def _get_azel2radec_quats(self):
        """Construct a sparsely sampled vector of Az/El->Ra/Dec quaternions.

        The interpolation times must be tied to the total observation so
        that the results do not change when data is distributed in time
        domain.

        """
        # One control point at least every 10 minutes.  Overkill but
        # costs nothing.
        n = max(2, 1 + int((self._lasttime - self._firsttime) / 600))
        times = np.linspace(self._firsttime, self._lasttime, n)
        quats = np.zeros([n, 4])
        for i, t in enumerate(times):
            quats[i] = self._get_coord_quat(t)
            # Make sure we have a consistent branch in the quaternions.
            # Otherwise we'll get interpolation issues.
            if i > 0 and (
                np.sum(np.abs(quats[i - 1] + quats[i]))
                < np.sum(np.abs(quats[i - 1] - quats[i]))
            ):
                quats[i] *= -1
        quats = qa.norm(quats)
        return times, quats

    @function_timer
    def _get_coord_quat(self, t):
        """Get the Az/El -> Ra/Dec conversion quaternion for boresight.

        We will apply atmospheric refraction and stellar aberration in
        the detector frame.

        """
        self._observer.date = self.to_DJD(t)
        # Set pressure to zero to disable atmospheric refraction.
        pressure = self._observer.pressure
        self._observer.pressure = 0
        # Rotate the X, Y and Z axes from horizontal to equatorial frame.
        # Strictly speaking, two coordinate axes would suffice but the
        # math is cleaner with three axes.
        #
        # PyEphem measures the azimuth East (clockwise) from North.
        # The direction is standard but opposite to ISO spherical coordinates.
        try:
            xra, xdec = self._observer.radec_of(0, 0, fixed=False)
            yra, ydec = self._observer.radec_of(-np.pi / 2, 0, fixed=False)
            zra, zdec = self._observer.radec_of(0, np.pi / 2, fixed=False)
        except Exception as e:
            # Modified pyephem not available.
            # Translated pointing will include stellar aberration.
            xra, xdec = self._observer.radec_of(0, 0)
            yra, ydec = self._observer.radec_of(-np.pi / 2, 0)
            zra, zdec = self._observer.radec_of(0, np.pi / 2)
        self._observer.pressure = pressure
        xvec, yvec, zvec = ang2vec(
            np.pi / 2 - np.array([xdec, ydec, zdec]), np.array([xra, yra, zra])
        )
        # Orthonormalize for numerical stability
        xvec /= np.sqrt(np.dot(xvec, xvec))
        yvec -= np.dot(xvec, yvec) * xvec
        yvec /= np.sqrt(np.dot(yvec, yvec))
        zvec -= np.dot(xvec, zvec) * xvec + np.dot(yvec, zvec) * yvec
        zvec /= np.sqrt(np.dot(zvec, zvec))
        # Solve for the quaternions from the transformed axes.
        X = (xvec[1] + yvec[0]) / 4
        Y = (xvec[2] + zvec[0]) / 4
        Z = (yvec[2] + zvec[1]) / 4
        d = np.sqrt(np.abs(Y * Z / X))  # Choose positive root
        c = d * X / Y
        b = X / c
        a = (xvec[1] / 2 - b * c) / d
        # qarray has the scalar part as the last index
        quat = qa.norm(np.array([b, c, d, a]))
        return quat

    @function_timer
    def free_azel_quats(self):
        self._boresight_azel = None
        self.cache.destroy("boresight_azel")

    @function_timer
    def free_radec_quats(self):
        self._boresight = None
        self.cache.destroy("boresight_radec")

    @function_timer
    def radec2quat(self, ra, dec, pa):
        qR = qa.rotation(ZAXIS, ra + np.pi / 2)
        qD = qa.rotation(XAXIS, np.pi / 2 - dec)
        qP = qa.rotation(ZAXIS, pa)  # FIXME: double-check this
        q = qa.mult(qR, qa.mult(qD, qP))

        if self._coord != "C":
            # Add the coordinate system rotation
            if self._coord == "G":
                q = qa.mult(quat_equ2gal, q)
            elif self._coord == "E":
                q = qa.mult(quat_equ2ecl, q)
            else:
                raise RuntimeError("Unknown coordinate system: {}".format(self._coord))
        return q

    def detoffset(self):
        return {d: np.asarray(self._fp[d]) for d in self._detlist}

    def _get(self, detector, start, n):
        # This class just returns data streams of zeros
        return np.zeros(n, dtype=np.float64)

    def _put(self, detector, start, data):
        raise RuntimeError("cannot write data to simulated data streams")
        return

    def _get_flags(self, detector, start, n):
        return np.zeros(n, dtype=np.uint8)

    def _put_flags(self, detector, start, flags):
        raise RuntimeError("cannot write flags to simulated data streams")
        return

    def _get_common_flags(self, start, n):
        return self._commonflags[start : start + n]

    def _put_common_flags(self, start, flags):
        raise RuntimeError("cannot write flags to simulated data streams")
        return

    def _get_hwp_angle(self, start, n):
        if self.cache.exists(self.HWP_ANGLE_NAME):
            angle = self.cache.reference(self.HWP_ANGLE_NAME)[start : start + n]
        else:
            angle = None
        return angle

    def _put_hwp_angle(self, start, flags):
        raise RuntimeError("cannot write HWP angle to simulated data streams")
        return

    def _get_times(self, start, n):
        return self._times[start : start + n]

    def _put_times(self, start, stamps):
        raise RuntimeError("cannot write timestamps to simulated data streams")
        return

    def _get_boresight(self, start, n, azel=False):
        if azel:
            if self._boresight_azel is None:
                raise RuntimeError("Boresight azel pointing was purged.")
            return self._boresight_azel[start : start + n]
        else:
            if self._boresight is None:
                raise RuntimeError("Boresight radec pointing was purged.")
            return self._boresight[start : start + n]

    def _get_boresight_azel(self, start, n):
        return self._get_boresight(start, n, azel=True)

    def _put_boresight(self, start, data):
        raise RuntimeError("cannot write boresight to simulated data streams")
        return

    def _put_boresight_azel(self, start, data):
        raise RuntimeError("cannot write boresight to simulated data streams")
        return

    @function_timer
    def read_boresight_az(self, local_start=0, n=0):
        """Read the boresight azimuth.

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            n (int): the number of samples to read.  If zero, read to end.

        Returns:
            (array): a numpy array containing the timestamps.
        """
        if n == 0:
            n = self.local_samples[1] - local_start
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot read boresight azimuth - process "
                "has no assigned local samples"
            )
        if (local_start < 0) or (local_start + n > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid".format(
                    local_start, local_start + n - 1
                )
            )
        return self._az[local_start : local_start + n]

    @function_timer
    def read_boresight_el(self, local_start=0, n=0):
        """Read the boresight elevation.

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            n (int): the number of samples to read.  If zero, read to end.

        Returns:
            (array): a numpy array containing the timestamps.
        """
        if n == 0:
            n = self.local_samples[1] - local_start
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot read boresight elevation - process "
                "has no assigned local samples"
            )
        if (local_start < 0) or (local_start + n > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid".format(
                    local_start, local_start + n - 1
                )
            )
        return self._el[local_start : local_start + n]

    @function_timer
    def _get_pntg(self, detector, start, n, azel=False):
        # FIXME: this is where we will apply atmospheric refraction and
        # stellar aberration corrections in the detector frame.  For
        # simulations they will only matter if we want to simulate the
        # error coming from ignoring them.
        boresight = self._get_boresight(start, n, azel=azel)
        detquat = self._fp[detector]
        return qa.mult(boresight, detquat)

    def _put_pntg(self, detector, start, data):
        raise RuntimeError("cannot write data to simulated pointing")
        return

    @function_timer
    def _get_position(self, start, n):
        # For this simple class, assume that the Earth is located
        # along the X axis at time == 0.0s.  We also just use the
        # mean values for distance and angular speed.  Classes for
        # real experiments should obviously use ephemeris data.
        rad = np.fmod((start - self._firsttime) * self._radpersec, 2.0 * np.pi)
        ang = self._radinc * np.arange(n, dtype=np.float64) + rad
        x = self._AU * np.cos(ang)
        y = self._AU * np.sin(ang)
        z = np.zeros_like(x)
        return np.ravel(np.column_stack((x, y, z))).reshape((-1, 3))

    def _put_position(self, start, pos):
        raise RuntimeError("cannot write data to simulated position")
        return

    @function_timer
    def _get_velocity(self, start, n):
        # For this simple class, assume that the Earth is located
        # along the X axis at time == 0.0s.  We also just use the
        # mean values for distance and angular speed.  Classes for
        # real experiments should obviously use ephemeris data.
        rad = np.fmod((start - self._firsttime) * self._radpersec, 2.0 * np.pi)
        ang = self._radinc * np.arange(n, dtype=np.float64) + rad + (0.5 * np.pi)
        x = self._earthspeed * np.cos(ang)
        y = self._earthspeed * np.sin(ang)
        z = np.zeros_like(x)
        return np.ravel(np.column_stack((x, y, z))).reshape((-1, 3))

    def _put_velocity(self, start, vel):
        raise RuntimeError("cannot write data to simulated velocity")
        return



import glob
import sys

import healpy as hp

run_type = sys.argv[1]

folder = "out_tiny_{}/".format(run_type)
ref = "ref_" + folder

# recursive glob requires Python 3.5
filenames = glob.glob(ref + "/**/*.fits", recursive=True)
assert len(filenames) > 0, "No FITS files in the reference folder"
for filename in filenames:

    print("Compare", filename)

    m = hp.read_map(filename.replace(ref, folder), verbose=False)
    m_ref = hp.read_map(filename, verbose=False)
    m_diff_std = (m - m_ref).std()

    assert m_diff_std < 1e-7, "Maps differ, std {}".format(m_diff_std)

print("Test passed")



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import argparse
import os

from ..timing import function_timer, Timer
from ..utils import Logger, Environment


def add_debug_args(parser):
    # `debug` may be already added
    try:
        parser.add_argument(
            "--debug",
            required=False,
            action="store_true",
            help="Enable extra debugging outputs",
            dest="debug",
        )
    except argparse.ArgumentError:
        pass
    try:
        parser.add_argument(
            "--no-debug",
            required=False,
            action="store_false",
            help="Disable extra debugging outputs",
            dest="debug",
        )
    except argparse.ArgumentError:
        pass
    parser.set_defaults(debug=False)
    return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os
import gc

import numpy as np

from ._libtoast import Environment, Timer, GlobalTimers, Logger

from ._libtoast import (
    AlignedI8,
    AlignedU8,
    AlignedI16,
    AlignedU16,
    AlignedI32,
    AlignedU32,
    AlignedI64,
    AlignedU64,
    AlignedF32,
    AlignedF64,
)

from ._libtoast import vsin, vcos, vsincos, vatan2, vsqrt, vrsqrt, vexp, vlog

from ._libtoast import (
    vfast_sin,
    vfast_cos,
    vfast_sincos,
    vfast_atan2,
    vfast_sqrt,
    vfast_rsqrt,
    vfast_exp,
    vfast_log,
    vfast_erfinv,
)

from .mpi import MPI, use_mpi


# This function sets the numba threading layer to (hopefully) be compatible with TOAST.
# The TOAST threading concurrency is used to attempt to set the numba threading.  We
# try to use the OpenMP backend for numba and then TBB.  The "workqueue" backend (which
# is process based).  May not be compatible with all systems, so we use that as a
# last resort.  This function should be called by any operators that use numba.

numba_threading_layer = None


def set_numba_threading():
    """Set the numba threading layer.

    For parallel numba jit blocks, the backend threading layer is selected at runtime
    based on an order set inside the numba package.  We would like to change the
    order of selection to prefer one of the thread-based backends (omp or tbb).  We also
    set the maximum number of threads used by numba to be the same as the number of
    threads used by TOAST.  Since TOAST does not use numba, it means that there will
    be a consistent maximum number of threads in use at all times and no
    oversubscription.

    Args:
        None

    Returns:
        None

    """
    global numba_threading_layer
    if numba_threading_layer is not None:
        # Already set.
        return

    # Get the number of threads used by TOAST at runtime.
    env = Environment.get()
    log = Logger.get()
    toastthreads = env.max_threads()
    print("max toast threads = ", toastthreads, flush=True)

    rank = 0
    if use_mpi:
        rank = MPI.COMM_WORLD.rank

    threading = "default"
    have_numba_omp = False
    try:
        # New style package layout
        from numba.np.ufunc import omppool

        have_numba_omp = True
        if rank == 0:
            log.debug("Numba has OpenMP threading support")
    except ImportError:
        try:
            # Old style
            from numba.npyufunc import omppool

            have_numba_omp = True
            if rank == 0:
                log.debug("Numba has OpenMP threading support")
        except ImportError:
            # no OpenMP support
            if rank == 0:
                log.debug("Numba does not support OpenMP")
    have_numba_tbb = False
    try:
        # New style package layout
        from numba.np.ufunc import tbbpool

        have_numba_tbb = True
        if rank == 0:
            log.debug("Numba has TBB threading support")
    except ImportError:
        try:
            # Old style
            from numba.npyufunc import tbbpool

            have_numba_tbb = True
            if rank == 0:
                log.debug("Numba has TBB threading support")
        except ImportError:
            # no TBB
            if rank == 0:
                log.debug("Numba does not support TBB")

    # Prefer OMP backend
    if have_numba_omp:
        threading = "omp"
    elif have_numba_tbb:
        threading = "tbb"

    try:
        from numba import vectorize, config, threading_layer

        # Set threading layer and number of threads.  Note that this still
        # does not always work.  The conf structure is repopulated from the
        # environment on every compilation if any of the NUMBA_* variables
        # have changed.
        config.THREADING_LAYER = threading
        config.NUMBA_DEFAULT_NUM_THREADS = toastthreads
        config.NUMBA_NUM_THREADS = toastthreads
        os.environ["NUMBA_THREADING_LAYER"] = threading
        os.environ["NUMBA_DEFAULT_NUM_THREADS"] = "{:d}".format(toastthreads)
        os.environ["NUMBA_NUM_THREADS"] = "{:d}".format(toastthreads)

        # In order to get numba to actually select a threading layer, we must
        # trigger compilation of a parallel function.
        @vectorize("float64(float64)", target="parallel")
        def force_thread_launch(x):
            return x + 1

        force_thread_launch(np.zeros(1))

        # Log the layer that was selected
        numba_threading_layer = threading_layer()
        if rank == 0:
            log.debug("Numba threading layer set to {}".format(numba_threading_layer))
            log.debug(
                "Numba max threads now forced to {}".format(config.NUMBA_NUM_THREADS)
            )
    except ImportError:
        # Numba not available at all
        if rank == 0:
            log.debug("Cannot import numba- ignoring threading layer.")


try:
    import psutil

    def memreport(msg="", comm=None):
        """Gather and report the amount of allocated, free and swapped system memory"""
        if psutil is None:
            return
        vmem = psutil.virtual_memory()._asdict()
        gc.collect()
        vmem2 = psutil.virtual_memory()._asdict()
        memstr = "Memory usage {}\n".format(msg)
        for key, value in vmem.items():
            value2 = vmem2[key]
            if comm is None:
                vlist = [value]
                vlist2 = [value2]
            else:
                vlist = comm.gather(value)
                vlist2 = comm.gather(value2)
            if comm is None or comm.rank == 0:
                vlist = np.array(vlist, dtype=np.float64)
                vlist2 = np.array(vlist2, dtype=np.float64)
                if key != "percent":
                    # From bytes to better units
                    if np.amax(vlist) < 2 ** 20:
                        vlist /= 2 ** 10
                        vlist2 /= 2 ** 10
                        unit = "kB"
                    elif np.amax(vlist) < 2 ** 30:
                        vlist /= 2 ** 20
                        vlist2 /= 2 ** 20
                        unit = "MB"
                    else:
                        vlist /= 2 ** 30
                        vlist2 /= 2 ** 30
                        unit = "GB"
                else:
                    unit = "% "
                if comm is None or comm.size == 1:
                    memstr += "{:>12} : {:8.3f} {}\n".format(key, vlist[0], unit)
                    if np.abs(vlist2[0] - vlist[0]) / vlist[0] > 1e-3:
                        memstr += "{:>12} : {:8.3f} {} (after GC)\n".format(
                            key, vlist2[0], unit
                        )
                else:
                    med1 = np.median(vlist)
                    memstr += (
                        "{:>12} : {:8.3f} {}  < {:8.3f} +- {:8.3f} {}  "
                        "< {:8.3f} {}\n".format(
                            key,
                            np.amin(vlist),
                            unit,
                            med1,
                            np.std(vlist),
                            unit,
                            np.amax(vlist),
                            unit,
                        )
                    )
                    med2 = np.median(vlist2)
                    if np.abs(med2 - med1) / med1 > 1e-3:
                        memstr += (
                            "{:>12} : {:8.3f} {}  < {:8.3f} +- {:8.3f} {}  "
                            "< {:8.3f} {} (after GC)\n".format(
                                key,
                                np.amin(vlist2),
                                unit,
                                med2,
                                np.std(vlist2),
                                unit,
                                np.amax(vlist2),
                                unit,
                            )
                        )
        if comm is None or comm.rank == 0:
            print(memstr, flush=True)
        if comm is not None:
            comm.Barrier()
        return


except ImportError:

    def memreport(msg="", comm=None):
        return


def object_ndim(x):
    """Get the number of dimension of an object.

    Scalars return 0.  Numpy arrays return their actual ndim value.
    Objects that support the buffer protocol return the ndim value of the
    corresponding memoryview.  Nested lists are traversed to compute the
    effective number of dimensions.

    Args:
        x (object): some stuff.

    Returns:
        (int): the number of dimensions of the stuff.

    """
    try:
        nd = x.ndim
        return nd
    except AttributeError:
        # Not a numpy array...
        try:
            view = memoryview(x)
            nd = view.ndim
            return nd
        except TypeError:
            # Does not support buffer protocol...
            try:
                lg = len(x)
                # It's a list!
                nd = 1
                cur = x[0]
                try:
                    lg = len(cur)
                    nd += 1
                    cur = cur[0]
                    try:
                        lg = len(cur)
                        nd += 1
                        # Using lists of more than 3 dimensions (rather than
                        # a numpy array) is kind of crazy...
                    except TypeError:
                        pass
                except TypeError:
                    pass
                return nd
            except TypeError:
                # Must be a scalar.
                return 0


def ensure_buffer_i64(data):
    """Return a flattened object that supports the buffer protocol.

    Numpy arrays and objects supporting the buffer protocol are flattened and
    copied to contiguous memory if needed.  Scalars and lists are converted
    to numpy arrays.

    Args:
        data (object):  A scalar or iterable.

    Returns:
        (array):  The input unmodified or converted to an array.

    """
    return np.ascontiguousarray(data, dtype=np.int64).flatten()


def ensure_buffer_f64(data):
    """Return a flattened object that supports the buffer protocol.

    Numpy arrays and objects supporting the buffer protocol are flattened and
    copied to contiguous memory if needed.  Scalars and lists are converted
    to numpy arrays.

    Args:
        data (object):  A scalar or iterable.

    Returns:
        (array):  The input unmodified or converted to an array.

    """
    return np.ascontiguousarray(data, dtype=np.float64).flatten()
    # if isinstance(data, np.ndarray):
    #     print("ensure: found numpy array, shape=", data.shape, flush=True)
    #     return np.ascontiguousarray(data.flatten(), dtype=np.float64)
    # try:
    #     view = memoryview(data)
    #     if view.ndim == 0:
    #         # A single element...
    #         print("ensure: converting scalar buffer", flush=True)
    #         return np.array([view], dtype=np.float64)
    #     elif (not view.c_contiguous) or (view.ndim != 1):
    #         print("ensure: found non-contiguous memory view shape=", view.shape, flush=True)
    #         return np.ascontiguousarray(view, dtype=np.float64)
    #     else:
    #         print("ensure: returning original data shape=", view.shape, flush=True)
    #         return data
    # except TypeError:
    #     # Does not support buffer protocol
    #     print("ensure: converting non-buffer object ", data, flush=True)
    #     return np.ascontiguousarray(data, dtype=np.float64)



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

from ..timing import function_timer

from ..op import Operator

from .._libtoast import cov_mult_diag, cov_apply_diag, cov_eigendecompose_diag

from .pixels import DistPixels


@function_timer
def covariance_invert(npp, threshold, rcond=None):
    """Invert a diagonal noise covariance.

    This does an inversion of the covariance.  The threshold is
    applied to the condition number of each block of the matrix.  Pixels
    failing the cut are set to zero.

    Args:
        npp (DistPixels): The distributed covariance.
        threshold (float): The condition number threshold to apply.
        rcond (DistPixels): (Optional) The distributed inverse condition number map
            to fill.

    Returns:
        None

    """
    mapnnz = int(((np.sqrt(8 * npp.nnz) - 1) / 2) + 0.5)
    nppdata = npp.flatdata
    if nppdata is None:
        nppdata = np.empty(shape=0, dtype=np.float64)
    if rcond is not None:
        if rcond.npix != npp.npix:
            raise RuntimeError(
                "covariance matrix and condition number map must have same number "
                "of pixels"
            )
        if rcond.npix_submap != npp.npix_submap:
            raise RuntimeError(
                "covariance matrix and condition number map must have same submap size"
            )
        if rcond.nnz != 1:
            raise RuntimeError("condition number map should have NNZ = 1")

        rdata = rcond.flatdata
        if rdata is None:
            rdata = np.empty(shape=0, dtype=np.float64)
        cov_eigendecompose_diag(
            npp.nsubmap, npp.npix_submap, mapnnz, nppdata, rdata, threshold, True
        )

    else:
        temp = np.zeros(shape=(npp.nsubmap * npp.npix_submap), dtype=np.float64)
        cov_eigendecompose_diag(
            npp.nsubmap, npp.npix_submap, mapnnz, nppdata, temp, threshold, True
        )
    return


@function_timer
def covariance_multiply(npp1, npp2):
    """Multiply two diagonal noise covariances.

    This does an in-place multiplication of the covariance.
    The data values of the first covariance (npp1) are replaced with
    the result.

    Args:
        npp1 (3D array): The first distributed covariance.
        npp2 (3D array): The second distributed covariance.

    Returns:
        None

    """
    mapnnz = int(((np.sqrt(8 * npp1.nnz) - 1) / 2) + 0.5)

    if npp1.npix != npp2.npix:
        raise RuntimeError("covariance matrices must have same number of pixels")
    if npp1.npix_submap != npp2.npix_submap:
        raise RuntimeError("covariance matrices must have same submap size")
    if npp1.nnz != npp2.nnz:
        raise RuntimeError("covariance matrices must have same NNZ values")

    npp1data = npp1.flatdata
    if npp1data is None:
        npp1data = np.empty(shape=0, dtype=np.float64)
    npp2data = npp2.flatdata
    if npp2data is None:
        npp2data = np.empty(shape=0, dtype=np.float64)
    cov_mult_diag(npp1.nsubmap, npp1.npix_submap, mapnnz, npp1data, npp2data)
    return


@function_timer
def covariance_apply(npp, m):
    """Multiply a map by a diagonal noise covariance.

    This does an in-place multiplication of the covariance and a
    map.  The results are returned in place of the input map.

    Args:
        npp (DistPixels): The distributed covariance.
        m (DistPixels): The distributed map.

    Returns:
        None

    """
    mapnnz = int(((np.sqrt(8 * npp.nnz) - 1) / 2) + 0.5)

    if m.npix != npp.npix:
        raise RuntimeError("covariance matrix and map must have same number of pixels")
    if m.npix_submap != npp.npix_submap:
        raise RuntimeError("covariance matrix and map must have same submap size")
    if m.nnz != mapnnz:
        raise RuntimeError("covariance matrix and map have incompatible NNZ values")

    nppdata = npp.flatdata
    if nppdata is None:
        nppdata = np.empty(shape=0, dtype=np.float64)
    mdata = m.flatdata
    if mdata is None:
        mdata = np.empty(shape=0, dtype=np.float64)
    cov_apply_diag(npp.nsubmap, npp.npix_submap, mapnnz, nppdata, mdata)
    return


@function_timer
def covariance_rcond(npp):
    """Compute the inverse condition number map.

    This computes the inverse condition number map of the supplied
    covariance matrix.

    Args:
        npp (DistPixels): The distributed covariance.

    Returns:
        rcond (DistPixels): The distributed inverse condition number map.
    """
    mapnnz = int(((np.sqrt(8 * npp.nnz) - 1) / 2) + 0.5)

    rcond = npp.duplicate(copy=False, nnz=1)

    threshold = np.finfo(np.float64).eps

    nppdata = npp.flatdata
    if nppdata is None:
        nppdata = np.empty(shape=0, dtype=np.float64)

    rdata = rcond.flatdata
    if rdata is None:
        rdata = np.empty(shape=0, dtype=np.float64)

    cov_eigendecompose_diag(
        npp.nsubmap, npp.npix_submap, mapnnz, nppdata, rdata, threshold, False
    )

    return rcond



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import re
import ctypes
import numpy as np

from .utils import (
    Logger,
    AlignedI8,
    AlignedU8,
    AlignedI16,
    AlignedU16,
    AlignedI32,
    AlignedU32,
    AlignedI64,
    AlignedU64,
    AlignedF32,
    AlignedF64,
)


class Cache(object):
    """Data cache with explicit memory management.

    This class acts as a dictionary of named arrays.  Each array may be
    multi-dimensional.

    Args:
        pymem (bool): if True, use python memory rather than external
            allocations in C.  Only used for testing.
    """

    def __init__(self, pymem=False):
        self._pymem = pymem
        self._buffers = dict()
        self._dtypes = dict()
        self._shapes = dict()
        self._aliases = dict()

    def __getitem__(self, key):
        return self.reference(key)

    def __setitem__(self, key, value):
        self.put(key, value, replace=True)

    def __delitem__(self, key):
        self.destroy(key)

    def __contains__(self, key):
        return self.exists(key)

    def __len__(self):
        return len(self.keys())

    def __iter__(self):
        class CacheIterator:
            def __init__(self, cache):
                self.cache = cache
                self.keys = cache.keys()

            def __iter__(self):
                return self

            def __next__(self):
                if len(self.keys) == 0:
                    raise StopIteration
                key = self.keys.pop()
                return self.cache[key]

        return CacheIterator(self)

    def clear(self, pattern=None):
        """Clear one or more buffers.

        Args:
            pattern (str): a regular expression to match against the buffer
                names when determining what should be cleared.  If None,
                then all buffers are cleared.

        Returns:
            None

        """
        if pattern is None:
            # free all buffers
            self._aliases.clear()
            self._buffers.clear()
            self._dtypes.clear()
            self._shapes.clear()
        else:
            pat = re.compile(pattern)
            names = list(self._buffers.keys())
            matching = list()
            for n in names:
                mat = pat.match(n)
                if mat is not None:
                    matching.append(n)
            for n in matching:
                self.destroy(n)
        return

    def create(self, name, type, shape):
        """Create a named data buffer of the given type and shape.

        Args:
            name (str): the name to assign to the buffer.
            type (numpy.dtype): one of the supported numpy types.
            shape (tuple): a tuple containing the shape of the buffer.

        Returns:
            (array): a reference to the allocated array.

        """
        log = Logger.get()
        if name is None:
            raise ValueError("Cache name cannot be None")
        if type is None:
            raise ValueError("Cache type cannot be None")
        if shape is None:
            raise ValueError("Cache shape cannot be None")
        if self.exists(name):
            raise RuntimeError("Data buffer or alias {} already exists".format(name))
        ttype = np.dtype(type)
        flatshape = 1
        for dim in shape:
            flatshape *= dim
        if self._pymem:
            self._buffers[name] = np.zeros(flatshape, dtype=ttype)
        else:
            if ttype.char == "b":
                self._buffers[name] = AlignedI8.zeros(flatshape)
            elif ttype.char == "B":
                self._buffers[name] = AlignedU8.zeros(flatshape)
            elif ttype.char == "h":
                self._buffers[name] = AlignedI16.zeros(flatshape)
            elif ttype.char == "H":
                self._buffers[name] = AlignedU16.zeros(flatshape)
            elif ttype.char == "i":
                self._buffers[name] = AlignedI32.zeros(flatshape)
            elif ttype.char == "I":
                self._buffers[name] = AlignedU32.zeros(flatshape)
            elif (ttype.char == "q") or (ttype.char == "l"):
                self._buffers[name] = AlignedI64.zeros(flatshape)
            elif (ttype.char == "Q") or (ttype.char == "L"):
                self._buffers[name] = AlignedU64.zeros(flatshape)
            elif ttype.char == "f":
                self._buffers[name] = AlignedF32.zeros(flatshape)
            elif ttype.char == "d":
                self._buffers[name] = AlignedF64.zeros(flatshape)
            else:
                msg = "Unsupported data typecode '{}'".format(ttype.char)
                log.error(msg)
                raise ValueError(msg)
        self._dtypes[name] = ttype
        self._shapes[name] = shape
        if self._pymem:
            return self._buffers[name].reshape(self._shapes[name])
        else:
            return self._buffers[name].array().reshape(self._shapes[name])
        return self._buffers[name]

    def put(self, name, data, replace=False):
        """Create a named data buffer to hold the provided data.

        If replace is True, existing buffer of the same name is first
        destroyed. If replace is True and the name is an alias, it is
        promoted to a new data buffer.

        Args:
            name (str): the name to assign to the buffer.
            data (numpy.ndarray): Numpy array
            replace (bool): Overwrite any existing keys

        Returns:
            (array): a numpy array wrapping the raw data buffer.

        """
        if name is None:
            raise ValueError("Cache name cannot be None")
        indata = data
        if self.exists(name):
            # This buffer already exists. Is the input data buffer actually
            # the same memory as the buffer already stored?  If so, just
            # return a new reference.
            realname = name
            if name in self._aliases:
                realname = self._aliases[name]
            addr = None
            if self._pymem:
                p_ref = self._buffers[realname].ctypes.data_as(ctypes.c_void_p).value
            else:
                p_ref = self._buffers[realname].address()

            p_data = data.ctypes.data_as(ctypes.c_void_p).value
            # print("p_ref = {}, p_data = {}".format(p_ref, p_data), flush=True)
            if (
                (p_ref == p_data)
                and (self._shapes[realname] == data.shape)
                and (self._dtypes[realname] == data.dtype)
            ):
                return self.reference(realname)
            if not replace:
                raise RuntimeError(
                    "Cache buffer named {} already exists "
                    "and replace is False.".format(name)
                )
            # At this point we have an existing memory buffer or alias with
            # the same name, and which is not identical to the input.  If this
            # is an alias, just delete it.
            if name in self._aliases:
                del self._aliases[name]
            else:
                # This existing data is not an alias.  However, the input
                # might be a view into this existing memory.  Before deleting
                # the existing data, we copy the input just in case.
                indata = np.array(data)
                self.destroy(name)

        # Now create the new buffer and copy in the data.
        ref = self.create(name, indata.dtype, indata.shape)
        np.copyto(ref, indata)
        return ref

    def add_alias(self, alias, name):
        """Add an alias to a name that already exists in the cache.

        Args:
            alias (str): alias to create
            name (str): an existing key in the cache

        Returns:
            None

        """
        if alias is None or name is None:
            raise ValueError("Cache name or alias cannot be None")
        names = list(self._buffers.keys())
        if name not in names:
            raise RuntimeError(
                "Data buffer {} does not exist for alias {}".format(name, alias)
            )
        if alias in names:
            raise RuntimeError(
                "Proposed alias {} would shadow existing buffer.".format(alias)
            )
        self._aliases[alias] = name
        return

    def destroy(self, name):
        """Deallocate the specified buffer.

        Only call this if all numpy arrays that reference the memory
        are out of use.  If the specified name is an alias, then the alias
        is simply deleted.  If the specified name is an actual buffer, then
        all aliases pointing to that buffer are also deleted.

        Args:
            name (str): the name of the buffer or alias to destroy.

        Returns:
            None

        """
        if name in self._aliases.keys():
            # Name is an alias. Do not remove the buffer
            del self._aliases[name]
            return
        names = list(self._buffers.keys())
        if name not in names:
            raise KeyError("Data buffer {} does not exist".format(name))

        # Remove aliases to the buffer
        aliases_to_remove = []
        for key, value in self._aliases.items():
            if value == name:
                aliases_to_remove.append(key)
        for key in aliases_to_remove:
            del self._aliases[key]

        # Forcibly resize this buffer to length zero
        if not self._pymem:
            self._buffers[name].clear()

        # Remove actual buffer
        del self._buffers[name]
        del self._dtypes[name]
        del self._shapes[name]
        return

    def exists(self, name):
        """Check whether a buffer exists.

        Args:
            name (str): the name of the buffer to search for.

        Returns:
            (bool):  True if a buffer or alias exists with the given name.

        """
        if name in self._aliases:
            # We have an alias with this name, so it exists.
            return True
        names = list(self._buffers.keys())
        if name in names:
            return True
        return False

    def reference(self, name):
        """Return a numpy array pointing to the buffer.

        The returned array will wrap a pointer to the raw buffer, but will
        not claim ownership.  When the numpy array is garbage collected, it
        will NOT attempt to free the memory (you must manually use the
        destroy method).

        Args:
            name (str): the name of the buffer to return.

        Returns:
            (array): a numpy array wrapping the raw data buffer.

        """
        # First check that it exists
        if not self.exists(name):
            raise KeyError("Data buffer (nor alias) {} does not exist".format(name))
        realname = name
        if name in self._aliases:
            # This is an alias
            realname = self._aliases[name]
        if self._pymem:
            return self._buffers[realname].reshape(self._shapes[realname])
        else:
            return self._buffers[realname].array().reshape(self._shapes[realname])

    def keys(self):
        """Return a list of all the keys in the cache.

        Returns:
            (list): List of key strings.

        """
        return sorted(list(self._buffers.keys()))

    def aliases(self):
        """Return a dictionary of all the aliases to keys in the cache.

        Returns:
            (dict): Dictionary of aliases.

        """
        return self._aliases.copy()

    def report(self, silent=False):
        """Report memory usage.

        Args:
            silent (bool):  Count and return the memory without printing.

        Returns:
            (int):  Amount of allocated memory in bytes

        """
        log = Logger.get()
        if not silent:
            log.info("Cache memory usage:")
        tot = 0
        for key in self.keys():
            ref = self.reference(key)
            sz = ref.nbytes
            del ref
            tot += sz
            if not silent:
                log.info(" - {:25} {:5.2f} MB".format(key, sz / 2 ** 20))
        if not silent:
            log.info(" {:27} {:5.2f} MB".format("TOTAL", tot / 2 ** 20))
        return tot
